""" data_frame module

Implements a extended pandas.DataFrame. It is based on the pandas.DataFrame and extended by many special methods that simplify working with electrical quantities.
This includes easy management of small signal parameter and other quantities which can be calculated from them.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import numpy as np
import re
import logging
import copy
from scipy.optimize import curve_fit
from DMT.exceptions import UnknownColumnError
from DMT.core import (
    specifiers_ss_para,
    get_specifier_from_string,
    specifiers,
    SpecifierStr,
    sub_specifiers,
    get_nodes,
    get_sub_specifiers,
    flatten,
    DataProcessor,
)
import pandas as pd

# pylint: disable = too-many-lines

# helper functions for Y Parameter smoothing
def fun_re(freq, a1, a2, a3):
    """Fit function for real Y Parameters normalized."""
    omega_sq = (2 * np.pi * freq) ** 2
    return (a1 + a2 * omega_sq) / (1 + a3 * omega_sq)


def fun_im(freq, a2, a3):
    """Fit function for imaginary Y Parameters normalized to omega."""
    omega = 2 * np.pi * freq
    omega_sq = omega**2
    return (a2) / (1 + a3 * omega_sq)


class Series(pd.Series):

    _metadata = ["processor"]

    @property
    def _constructor(self):
        return Series

    @property
    def _constructor_expanddim(self):
        return DataFrame


class DataFrame(DataProcessor, pd.DataFrame):
    """Two-dimensional size-mutable, potentially heterogeneous tabular data structure for electrical data with labeled axes (rows and columns).

    The primary DMT data structure. This class inherits from pandas.DataFrame and from the Mixin class DataProcessor.

    The inheritance from pandas is implemented according `to <http://pandas.pydata.org/pandas-docs/stable/extending.html#extending-subclassing-pandas>`.

    The inheritance from Mixin class DataProcessor is implemented according `to <http://www.qtrac.eu/pyagg.html>`.

    Note that pandas DataFrames are not always mutable and hence class methods that modify the DataFrame should be called like this::

        df = df.calc_ft()

    Methods
    -------
    convert_n_port_para(p_from='', p_to='', z0=float(50), ports=None)
        convert the small signal parameters of type p_from to small signal parameters of type p_to.
    cmplx2real()
        find all complex columns in self and split them up into one column for the real and one column for the complex part.
    real2cmplx()
        find all associated columns that contain real and imaginary parts and merge them into one column that holds complex numbers.
    clean_data(nodes, fallback=None, specifier_voltage=specifiers.VOLTAGE, specifier_current=specifiers.CURRENT, specifier_capacitance=specifiers.CAPACITANCE, specifier_frequency=specifiers.FREQUENCY, specifier_temperature=specifiers.TEMPERATURE)
        convert a DataFrame into the DMT format, e.g. use specifiers for the column names, drop unknown or unncessary columns and so on.
    clean_names( specifier_voltage=specifiers.VOLTAGE, specifier_current=specifiers.CURRENT, specifier_capacitance=specifiers.CAPACITANCE, specifiers_ss_parameter=specifiers.SPECIFIERS_SS_PARA)
        convert all column names into a nicer format.
    get_col_name(specifier, *nodes, sub_specifiers='')
        get the content of the column specified.y by the input to this method.
    ensure_specifier_column(specifier, *nodes, sub_specifiers=None, ports=None)
        make sure that the specifier specified by the arguments exists in the DataFrame.
    create_voltage(nodes, sub_specifiers=None)
        Try to create the voltage between nodes and save it into self.
    create_potential(nodes, sub_specifiers=None)
        Try to create the potential at node and save it into self.
    drop_all_voltages()
        Delete all voltages (but keep potentials).
    get_all_voltages()
        Return all voltages.
    parallel_norm(n_parallel, port_1, port_2)
        Normalize Currents and small signal parameters assuming that n_parallel devices have been measured.
    deembed_short(df_short, ports)
        perform a short deembeding of self with df_short.
    deembed_open(df_open, ports)
        perform an open deembeding of self with df_short.
    deembed(df_open, df_short, ports=None)
        perform open short deembeding of self with df_short and df_open.
    determine_mres()
        assume that self contains measurements of short structures. determine the metallization resistances using polyfits.
    deembed_DC(df_short_dc)
        perform DC deembeding of the currents in self, assuming that the metallization resistances can be calculated from df_short_dc.
    check_ss_cols(para)
        check if the small signal parameters para are existent in self.
    get_ss_para(para, port_1, *ports_n)
        return the small signal parameter in self.
    set_ss_para(para, para_values, port_1, *ports_n)
        set the small signal parameters para to self.
    strip_ss_para(keep='S')
        remove all small signal parameters except those specified by the keep argument.
    get_all_ss_para()
        return all small signal parameters in self.
    calc_ft(port_1, port_2)
        calculate ft using the spot frequency method.
    calc_fmax(port_1, port_2)
        calculate fmax from msg.
    calc_msg(port_1, port_2)
        calculate msg and save into self.
    calc_unilateral_gain(port_1, port_2)
        calculate u and save into self.
    calc_mag(port_1, port_2)
        calculate mag and save into self.
    calc_k(port_1, port_2)
        calculate k and save into self.
    calc_cbe()
        calculate cbe and save into self.
    calc_cbc()
        calculate cbc and save into self.
    calc_beta()
        calculate beta and save into self.
    calc_gm()
        calculate gm and save into self.
    """

    _metadata = ["processor"]
    processor = DataProcessor()

    @property
    def _constructor(self):
        DataFrame.processor = DataProcessor()
        return DataFrame

    @property
    def _constructor_sliced(self):
        Series.processor = DataProcessor()
        return Series

    def repeat_rows(self, count):
        """Repeats the DataFrame rows by the number count.

        Parameters
        ----------
        count : int
            Number of repeats per row.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe with repeated rows
        """
        data_new = []
        for row in self.itertuples(index=False, name=None):
            for _nr in range(count):
                data_new.append(row)

        df_new = DataFrame(columns=self.columns, data=data_new)

        return df_new

    def convert_n_port_para(self, p_from="", p_to="", z0=float(50), ports=None):
        """Convert between the small signal parameters S,Y,Z,T,A stored in self.

        Parameters
        ----------
        p_from  :  string
            String that specifies which parameters are already present in self.

        p_to    :  string
            String that specifies which parameters should be created in self.

        z0      : float()
            Reference impedance in ohms, default is float(50).

        ports   : [str], optional
            Name of the ports in correct order. Defaults to ['1', '2'], for a BJT in common-emitter this would be ['B', 'C'].

        Returns
        -------
        self       :  :class:`DMT.core.DataFrame`
            Dataframe that contains the p_to n_port parameters
        """
        if ports is None:
            ports = ["1", "2"]

        # find the number of ports from available columns in df
        para = self.get_ss_para(p_from, ports[0], *ports[1:])
        para_new = self.processor.convert_n_port_para(para, p_from, p_to)

        # save the new parameters into the df
        return self.set_ss_para(p_to, para_new, ports[0], *ports[1:])

    def cmplx2real(self):
        """Create real and imaginary columns from complex columns in DataFrame.

        Returns
        -------
        self       :  :class:`DMT.core.DataFrame`
            Dataframe that contains real and imag columns instead of cmplx columns.

        """
        # find complex variables
        cmplx_cols = []
        for column in self.columns:  # pylint: disable=not-an-iterable
            # this column contains an explicit real part
            if self[column].to_numpy().dtype in ["cmplx128", np.dtype("complex128")]:
                cmplx_cols.append(column)

        # delete unncessary variables
        for var in cmplx_cols:
            self["R:" + var] = np.real(self[var].to_numpy())
            self["I:" + var] = np.imag(self[var].to_numpy())
            self = self.drop(columns=[var])

        return self

    def real2cmplx(self):
        """Add real and imaginary columns in DataFrame to complex columns.

        Returns
        -------
        self :  :class:`DMT.core.DataFrame`
            Dataframe that contains cmplx columns instead of seperate real and imag columns.

        """
        # find complex variables
        cmplx_vars_old, cmplx_vars_speci = [], []
        for column in self.columns:  # pylint: disable=not-an-iterable
            # this column contains an explicit real part
            if "R:" in column:
                cmplx_vars_old.append(column[2:])
            if sub_specifiers.REAL in column:
                cmplx_vars_speci.append(column[:-5])

        # delete unnecessary variables
        # vairables in oldschool grammar
        for var in cmplx_vars_old:
            pd.options.mode.chained_assignment = None  # default='warn'
            self[var] = self["R:" + var].to_numpy() + 1j * self["I:" + var].to_numpy()
            self = self.drop(columns=["R:" + var])
            self = self.drop(columns=["I:" + var])

        # variables in specifier grammar
        for var in cmplx_vars_speci:
            pd.options.mode.chained_assignment = None  # default='warn'
            try:  # sometimes we have mixed AC specifiers (e.g. Y_CB, Y_CB|REAL but no Y_CB|IMAG), then this will fail. but it is ok.
                self[var] = (
                    self[var + sub_specifiers.REAL].to_numpy()
                    + 1j * self[var + sub_specifiers.IMAG].to_numpy()
                )
            except KeyError:
                pass
            try:
                self = self.drop(columns=[var + sub_specifiers.REAL])
            except KeyError:
                pass
            try:
                self = self.drop(columns=[var + sub_specifiers.IMAG])
            except KeyError:
                pass

        return self

    def clean_data(
        self,
        nodes,
        reference_node,
        fallback=None,
        ac_ports=None,
        specifier_voltage=specifiers.VOLTAGE,
        specifier_current=specifiers.CURRENT,
        specifier_capacitance=specifiers.CAPACITANCE,
        specifier_frequency=specifiers.FREQUENCY,
        specifier_temperature=specifiers.TEMPERATURE,
        warnings=True,
    ):
        """Clean DataFrame.  This is one of the most important methods and should be called on all imported data. This ensures data consistency within DMT!

        Clean a DataFrame. This means:

        * make sure all potentials and voltages of the dut are stored correctly
        * convert separate real and imaginary parts of complex numbers into a single cmplx128 number
        * assign a :class:`~DMT.core.specifiers.SpecifierStr` as column name if possible.

        If the nodes in the columns do not match with the nodes in the nodes parameter, the fallback dictionary is used to try and rename the nodes.


        Parameters
        ----------
        nodes  :  [str]
            List of strings with the node names.
        fallback  :  {string:string, specifier_frequency:'FREQ', specifier_temperature='TEMP'}
            A dict that can be used to remap nodes in the file to nodes parameter. The keys of this dict are
            node names that may be errorneous or duplicate in the DataFrame and the values are the actual correct node
            names, according to DMT specifiers. E.g. one may have a voltage V_C1 in the DataFrame, but it should actually be V_C. Then fallback={'C1':'C'}
            will save the day. So the key is the current column name and the value the desired column name.
            The specifier_frequency and specifier_temperature are set to this dictionary.
            If the value of a key in this dict is None, the column will be dropped.

        ac_ports : [str]
            List that specifies the connection of the AC ports. E.g. ['B','C'] for common emitter.
        specifier_voltage : str
            Specifier for the voltage, defaults to 'V'
        specifier_current : str
            Specifier for the voltage, defaults to 'I'
        specifier_capacitance : str
            Specifier for the voltage, defaults to 'C'
        specifier_frequency : str
            Specifier for the frequency, defaults to 'FREQ'
        specifier_temperature : str
            Specifier for the column names in the measurements. Can be used to identify variables which have different names in the measurements.
            Inside of DMT the given default names are assumed to be valid and hence they can be renamed using these specifiers.
            Voltages, currents and capacitances are renamed in :meth:`~DMT.core.data_frame.DataFrame.clean_names`,
            temperature and frequency are set as fallbacks for :meth:`~DMT.core.naming.get_nodes`, defaults to 'TEMP'

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains columns according to the DMT internal standard.

        Examples
        --------

        >>> df.columns
        ['V_BE','V_c','V_c_','I_B']
        >>> df.clean_data('B', 'E', {'c_':'C'})
        >>> df.columns
        ['V_B','V_E','V_C','I_B'] # the second V_c_ got deleted in this case

        Notes
        -----
        ..todo: fall also needs to be cast to upper()
        """
        # ensure cmplx variables
        self = self.real2cmplx()

        if fallback is None:
            fallback = {}

        # clean the notation
        self = self.clean_names(
            specifier_voltage=specifier_voltage,
            specifier_current=specifier_current,
            specifier_capacitance=specifier_capacitance,
            ignore=fallback.keys(),
        )

        if fallback is None:
            fallback = {}
        fallback[specifier_temperature] = specifiers.TEMPERATURE
        fallback[specifier_frequency] = specifiers.FREQUENCY

        # rename columns that are specified by fallback
        new_df = DataFrame()
        unknown_columns = []
        for col in self.columns:
            if col in fallback.keys():
                # the complete column name is in the fallback. So use it!
                if fallback[col] is None:
                    if warnings:
                        logging.warning(
                            "The column %s is dropped as it is in the fallback dictionary!", col
                        )
                    # drop it!
                elif fallback[col]:
                    if warnings:
                        logging.warning(
                            "The column %s is kept as %s according to the fallback dictionary!",
                            col,
                            fallback[col],
                        )
                    new_df[fallback[col]] = self[col]
                else:
                    if warnings:
                        logging.warning(
                            "The column %s is kept it was according to the fallback dictionary!",
                            col,
                        )
                    new_df[col] = self[col]

                continue

            if re.search(r"deemb", col, re.IGNORECASE):
                raise IOError(
                    "Found a possible already deembeded column. This column must be dropped for DMT! To do that add the following to the fallback: {{'{0:s}': None}}".format(
                        col
                    )
                )

            if col[0] in specifiers_ss_para:
                if "1" in fallback and "2" in fallback:  # already in fallback ?
                    pass
                try:
                    fallback["1"] = ac_ports[0]
                    fallback["2"] = ac_ports[1]
                except KeyError:
                    raise IOError(
                        "DMT->DataFrame->clean_data: The column"
                        + str(col[0])
                        + " was identified as a small-signal parameter but the ac_ports argument was not specified."
                    )

            elif col[0] not in "VIC":
                # only rename the names of values which can be identified
                new_df[col] = self[col]
                continue

            nodes_in_col = get_nodes(col, nodes, fallback=fallback)
            sub_specifiers_in_col = get_sub_specifiers(col)
            # if more than 3 nodes, something is weird and we dont do anything
            # if len(nodes_in_col)>2:
            #     # only rename the names of values which can be identified
            #     new_df[col] = self[col]
            #     continue
            if nodes_in_col is None:
                continue
            elif nodes_in_col:
                # new_column_name = col[0] + r'_' + ''.join(nodes_in_col)
                new_column_name = SpecifierStr(
                    col[0], *nodes_in_col, sub_specifiers=sub_specifiers_in_col
                )
                # self  =  self.rename(columns={ self.columns[i]:new_column_name })
                if new_column_name in new_df.columns:  # pylint: disable=unsupported-membership-test
                    raise IOError(
                        "Column is already in the dataframe. Maybe it should be deleted? (fallback: {"
                        + new_column_name
                        + " :None})"
                    )
                new_df[new_column_name] = self[col]
            else:
                unknown_columns.append(col)

        if len(unknown_columns) > 1:
            str_columns = ""
            for col in unknown_columns:
                str_columns += " " + str(col) + " "

            raise UnknownColumnError(
                "The columns "
                + str_columns
                + " is unknown to this DuT as the nodes can not be extracted! Either a fallback behavior or different DuT nodes for this column are needed."
            )

        self = new_df

        # convert voltages, potentials and maybe in the future to SpecifierStr
        to_rename = {}
        temp_nodes = nodes + [
            key for key, value in fallback.items() if value == ""
        ]  # for this case also the "keep"-nodes from the fallback are possible
        for column in self.columns:  # pylint: disable=not-an-iterable
            specifier_col = get_specifier_from_string(column, temp_nodes)
            to_rename[column] = specifier_col

        self = self.rename(index=str, columns=to_rename)

        # set reference potential
        self.create_potential([reference_node], reference_node)

        # find all existing potentials and voltages and clean their name
        potentials = []
        voltages = []
        for column in self.columns:
            try:
                if column.specifier == specifiers.VOLTAGE:
                    # decide wether this colum contains a potential or a voltage
                    if len(column.nodes) == 1:
                        potentials.append(column)
                    elif len(column.nodes) == 2:
                        voltages.append(column)
                    elif len(column.nodes) == 0:
                        logging.warning(
                            "DMT -> Data_manager: Encountered voltage %s that corresponds to an unkown node. Skipping.",
                            column,
                        )
                        continue
                    else:
                        raise OSError(
                            "error: found voltage with name "
                            + column
                            + " in .mdm file that contains more than two node names!"
                        )

            # only works with specifiers
            except AttributeError:
                pass

        # get the potentials (nodes are known at this point), voltages later if-needed
        # find missing potentials
        possible_potentials = [SpecifierStr(specifiers.VOLTAGE, node) for node in nodes]
        missing_potentials = [
            potential for potential in possible_potentials if potential not in potentials
        ]
        if len(missing_potentials) == 0 or len(potentials) == 0:
            return self  # probably no voltage specified...

        # try to create all potentials that can be deduced from the voltages:
        for potential in missing_potentials:
            # find voltage that contains the potential's node
            try:
                # sometimes voltages with sub_specifiers are present in the data. We don't use these voltages here.
                self.create_potential(potential.nodes, reference_node)
            except IOError:
                # sometimes it may be impossible to create the potentials this way.
                # Assume e.g. V_S is missing and we dont have any voltage that contains the S node
                # Then we will just live with a missing potential.
                continue

        # remove the voltages
        self = self.drop(columns=voltages)

        return self

    # pylint: disable=dangerous-default-value
    def clean_names(
        self,
        specifier_voltage: str = specifiers.VOLTAGE,
        specifier_current: str = specifiers.CURRENT,
        specifier_capacitance: str = specifiers.CAPACITANCE,
        specifiers_ss_parameter=specifiers_ss_para,
        ignore: list = None,
    ):
        r"""Clean column names of DataFrame.

        Clean the column names of DataFrame into the DMT standard:

        * Goal is to have all column names as a :class:`~DMT.core.specifiers.SpecifierStr`.
        * potentials and voltages start with 'V\_'
        * currents start with 'I\_'
        * capacitances start with 'C\_'

        Specifier for the column names in the measurements. Can be used to identify variables which have different names in the measurements.
        Inside of DMT the given default names are assumed to be valid and hence they can be renamed using these specifiers.


        Parameters
        ----------
        specifier_voltage : str, optional
            Indicator for voltages in the measurements, by default specifiers.VOLTAGE
        specifier_current : str, optional
            Indicator for currents in the measurements, by default specifiers.CURRENT
        specifier_capacitance : str, optional
            Indicator for capacitances in the measurements, by default specifiers.CAPACITANCE
        specifiers_ss_parameter : optional
            Small signal parameters in the measurements, by default specifiers_ss_para
        ignore : list, optional
            Columns in the measurements to ignore, by default None

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains columns with names according to the DMT internal standard.
        """
        if ignore is None:
            ignore = []

        for i, column in enumerate(self.columns):
            if column in ignore:
                continue

            column = column.upper()
            try:
                if not column[1] == ":":  # Not an real or imaginary part of something
                    if column.startswith(specifier_voltage):  # is this a potential or voltage?
                        if column[1] != "_":
                            column = "V_" + column[1:]
                        else:
                            column = "V" + column[1:]
                    elif column.startswith(specifier_current):  # is this a current
                        if column[1] != "_":
                            column = "I_" + column[1:]
                        else:
                            column = "I" + column[1:]
                    elif column.startswith(specifier_capacitance):  # is this a capacitance
                        if column[1] != "_":
                            column = "C_" + column[1:]
                        else:
                            column = "C" + column[1:]
                    elif re.search(r"[(]([0-9]+),([0-9]+)[)]", column):
                        if column[0] in specifiers_ss_parameter and column[1] == "(":
                            ports = re.search(r"[(]([0-9]+),([0-9]+)[)]", column)

                            column = SpecifierStr(column[0], *[ports.group(1), ports.group(2)])

                self = self.rename(
                    columns={self.columns[i]: column}
                )  # pylint: disable=unsubscriptable-object

            except IndexError:
                pass

        return self

    def get_col_name(self, specifier, *nodes, sub_specifiers=""):
        """Checks if a column with the given combination of specifier and nodes exists in the given dataframe using the DMT naming style.

        Parameters
        ----------
        specifier : str
            Ideally a specifier from :mod:`~DMT.core.specifiers`
        *nodes : str, optional
            Nodes for the column. Like 'B' for base or 'B', 'E' for base-emitter
        sub_specifiers : str, {'', :py:const:`DMT.core.sub_specifiers.PERIMETER`, :py:const:`DMT.core.sub_specifiers.AREA`}

        Returns
        -------
        str
            If the column exists the name is returned.

        Raises
        ------
        KeyError
            If the column does not exist.
        """
        col = SpecifierStr(specifier, *nodes, sub_specifiers=sub_specifiers)
        if str(col) in self.columns:  # pylint: disable=unsupported-membership-test
            return col

        raise KeyError("The column '" + str(col) + "' is missing in the given data frame.")

    def ensure_specifier_column(
        self,
        specifier,
        *nodes,
        sub_specifiers_to_ensure=None,
        ports=None,
        reference_node=None,
        debug=False,
        area=None,
        force=False,
    ):
        """Checks if a column with the specifier name exists in the dataframe, if not it tries to calculate it!

        Parameters
        ----------
        specifier : str
            Ideally a specifier from :mod:`~DMT.core.specifiers`
        *nodes : str, optional
            Nodes for the column. Like 'B' for base or 'B', 'E' for base-emitter
        sub_specifiers : str, {'', :py:const:`DMT.core.sub_specifiers.PERIMETER`, :py:const:`DMT.core.sub_specifiers.AREA`}
            Subscript of the specifier symbol to ensure.
        ports        : [str], optional
            Name of the ports, for a BJT in common-emitter this is ['B', 'C', ...] . Only needed to calculate AC-Quantities.
        reference_node : str, optional
            Name of the common reference node, for a BJT in common-emitter this is 'E' . Only needed to calculate potentials.
        debug : bool, False
            If True, debug info is printed.
        area : float, None
            Specific area used to calculate area specific quantities.
        force : Bool, False
            If true: re-calculate specifier, even if already existing.

        Raises
        ------
        KeyError
            If the column does not exist.
        """
        if not self.columns.is_unique:
            self = self.loc[:, ~self.columns.duplicated()]
        col = SpecifierStr(specifier, *nodes, sub_specifiers=sub_specifiers_to_ensure)
        if col in self.columns and not force:  # pylint: disable=unsupported-membership-test
            return

        specifier = col.specifier
        nodes = col.nodes
        sub_specifiers_in_col = col.sub_specifiers

        if (
            sub_specifiers.REAL.sub_specifiers <= sub_specifiers_in_col
        ):  # catch them here before we enter the switch case below. This way we do not need to care about this below.
            col_complex = SpecifierStr(
                specifier,
                *nodes,
                sub_specifiers=sub_specifiers_in_col - sub_specifiers.REAL.sub_specifiers,
            )
            self.ensure_specifier_column(col_complex, ports=ports)
            self[col] = np.real(self[col_complex].to_numpy())

        elif (
            sub_specifiers.IMAG.sub_specifiers <= sub_specifiers_in_col
        ):  # [1:] to cut off the | for a correct string compare
            col_complex = SpecifierStr(
                specifier,
                *nodes,
                sub_specifiers=sub_specifiers_in_col - sub_specifiers.IMAG.sub_specifiers,
            )
            self.ensure_specifier_column(col_complex, ports=ports)
            self[col] = np.imag(self[col_complex].to_numpy())
            return
        elif (
            sub_specifiers.MAG.sub_specifiers <= sub_specifiers_in_col
        ):  # [1:] to cut off the | for a correct string compare
            col_complex = SpecifierStr(
                specifier,
                *nodes,
                sub_specifiers=sub_specifiers_in_col - sub_specifiers.MAG.sub_specifiers,
            )
            self.ensure_specifier_column(col_complex, ports=ports)
            self[col] = np.abs(self[col_complex].to_numpy())

        elif (
            sub_specifiers.PHASE.sub_specifiers <= sub_specifiers_in_col
        ):  # [1:] to cut off the | for a correct string compare
            col_complex = SpecifierStr(
                specifier,
                *nodes,
                sub_specifiers=sub_specifiers_in_col - sub_specifiers.PHASE.sub_specifiers,
            )
            self.ensure_specifier_column(col_complex, ports=ports)
            self[col] = np.angle(self[col_complex].to_numpy())
            return

        if (specifier == specifiers.VOLTAGE) and (len(nodes) == 2):
            try:
                self = self.create_voltage(
                    nodes,
                    reference_node=reference_node,
                    voltage_sub_specifiers=sub_specifiers_in_col,
                )
            except IOError as err:
                raise KeyError(
                    "The voltage '"
                    + str(col)
                    + "' and the needed potentials are missing in the given data frame and can not be calculated."
                ) from err
        elif (specifier == specifiers.VOLTAGE) and (len(nodes) == 1):
            try:
                self = self.create_potential(
                    nodes, reference_node, voltage_sub_specifiers=sub_specifiers_in_col, debug=debug
                )
            except IOError as err:
                raise KeyError(
                    "The potential '"
                    + str(col)
                    + "' and the needed voltages are missing in the given data frame and can not be calculated."
                ) from err

        elif specifier == specifiers.CURRENT_DENSITY:
            if area is None:
                raise IOError(
                    "DMT -> DataFrame -> ensure_specifier_column: area not given, but tried to calcualte current DENSITY. Abort."
                )
            self.loc[:, specifiers.CURRENT_DENSITY + nodes[0]] = (
                self[specifiers.CURRENT + nodes[0]] / area
            )
        elif (specifier == specifiers.CAPACITANCE) and (len(nodes) == 2):
            if ports is None:
                raise IOError(
                    'DMT -> DataFrame -> ensure_specifier_column: Calculation of a capacitance requires the specification of the "ports" keyword argument.'
                )
            if any(
                sub_specifier_poa.sub_specifiers <= sub_specifiers_in_col
                for sub_specifier_poa in [
                    sub_specifiers.PERIMETER,
                    sub_specifiers.AREA,
                    sub_specifiers.CORNER,
                    sub_specifiers.LENGTH,
                    sub_specifiers.WIDTH,
                ]
            ):
                raise IOError(
                    "DMT -> DataFrame -> ensure_specifier_column: Can not calculate a PoA capacitance. This needs to be done by a XQ Step."
                )

            try:
                if nodes[0] == "B" and nodes[1] == "E":  # CBE
                    self = self.calc_cbe(port_1=ports[0], port_2=ports[1])
                elif nodes[0] == "C" and nodes[1] == "E":  # CCE
                    self = self.calc_cce(port_1=ports[0], port_2=ports[1])
                elif nodes[0] == "B" and nodes[1] == "C":  # CBC
                    self = self.calc_cbc(port_1=ports[0], port_2=ports[1])
                else:
                    raise KeyError("The " + "".join(nodes) + " capacitance can not be calculated.")
            except IOError as err:
                raise KeyError(
                    "The " + "".join(nodes) + " capacitance could not be calculated."
                ) from err

        elif specifier == specifiers.TRANSIT_FREQUENCY:
            try:
                self = self.calc_ft(ports[0], ports[1])
            except IOError as err:
                raise KeyError(
                    "The transit frequency is missing in the given data frame and can not be calculated."
                ) from err
            except TypeError as err:
                raise IOError(
                    "The ac ports are not given, but transit frequency is beeing requested. Abort."
                ) from err

        elif specifier == specifiers.MAXIMUM_OSCILLATION_FREQUENCY:
            try:
                self = self.calc_fmax(ports[0], ports[1])
            except IOError as err:
                raise KeyError(
                    "Fmax is missing in the given data frame and can not be calculated."
                ) from err

        elif specifier == specifiers.MAXIMUM_AVAILABLE_GAIN:
            try:
                self = self.calc_mag(ports[0], ports[1])
            except IOError as err:
                raise KeyError(
                    "MAG is missing in the given data frame and can not be calculated."
                ) from err

        elif specifier == specifiers.MAXIMUM_STABLE_GAIN:
            try:
                self = self.calc_msg(ports[0], ports[1])
            except IOError as err:
                raise KeyError(
                    "MSG is missing in the given data frame and can not be calculated."
                ) from err

        elif specifier == specifiers.UNILATERAL_GAIN:
            try:
                self = self.calc_unilateral_gain(ports[0], ports[1])
            except IOError as err:
                raise KeyError(
                    "U is missing in the given data frame and can not be calculated."
                ) from err

        elif specifier == specifiers.TRANSIT_TIME:
            try:
                self.ensure_specifier_column(specifiers.TRANSIT_FREQUENCY, ports=ports)
                self[specifiers.TRANSIT_TIME] = 1 / (2 * np.pi * self[specifiers.TRANSIT_FREQUENCY])
            except IOError as err:
                raise KeyError(
                    "The transit time is missing in the given data frame and can not be calculated."
                ) from err
        elif specifier == specifiers.DC_CURRENT_AMPLIFICATION:
            try:
                self = self.calc_beta()
            except IOError as err:
                raise KeyError(
                    "The dc current amplification beta is missing in the given data frame and can not be calculated."
                ) from err
        elif specifier == specifiers.TRANSCONDUCTANCE:
            try:
                self = self.calc_gm()
            except IOError as err:
                raise KeyError(
                    "The transconductance is missing in the given data frame and can not be calculated."
                ) from err
        elif specifier in specifiers_ss_para:
            try:  # try tog et from S paras
                self = self.convert_n_port_para(p_from="S", p_to=specifier, ports=ports)
            except IOError as err:  # try to get from Y paras
                try:
                    self = self.convert_n_port_para(p_from="Y", p_to=specifier, ports=ports)
                except:
                    raise KeyError(
                        "The conversion from the S- and Y-Parameters to the small signal "
                        + specifier
                        + "-Parameters is missing.\n If possible add the conversion in the data processor!"
                    ) from err
            except KeyError as err:
                if ports is None:
                    raise KeyError(
                        "The small signal parameters "
                        + specifier
                        + "- and the S-Parameters are missing in the DataFrame.",
                        "Did you forget to pass the ports as ports=['B', 'C'] ?",
                    ) from err
                else:
                    raise KeyError(
                        "The small signal parameters "
                        + specifier
                        + "- and the S-Parameters are missing in the DataFrame."
                    ) from err

        else:
            raise KeyError(
                "The column '"
                + str(col)
                + "' is missing in the given data frame and no calculation for this column is implemented."
            )
        if not self.columns.is_unique:
            self = self.loc[:, ~self.columns.duplicated()]

    def create_voltage(self, nodes, voltage_sub_specifiers=None, reference_node=None):
        """Try to create col from existing data in dataframe df.

        Parameters
        ----------
        nodes  :  [node]
            List containing the nodes present in self or just directly list of nodes.
        sub_specifiers : [str]
            The sub specifier that shall be valid for the voltage to be created by this method.
        reference_node : str
            Reference node of the potentials, defaults to 'E'

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DMT.Dataframe with new dataframe, if successfull
        """
        if len(nodes) != 2:
            raise IOError(
                "DMT -> Data_manager: Tried to create column voltage but could not identify suitable potentials."
            )

        # create specifier for the new voltage
        col = SpecifierStr(specifiers.VOLTAGE, *nodes, sub_specifiers=voltage_sub_specifiers)

        # init specifiers for the potentials
        potential_0 = SpecifierStr(
            specifiers.VOLTAGE, nodes[0], sub_specifiers=voltage_sub_specifiers
        )
        potential_1 = SpecifierStr(
            specifiers.VOLTAGE, nodes[1], sub_specifiers=voltage_sub_specifiers
        )

        # ensure the existence of potentials
        self.ensure_specifier_column(potential_0, reference_node=reference_node)
        self.ensure_specifier_column(potential_1, reference_node=reference_node)

        # I think the setting with Copy Warning is broken here...
        pd.options.mode.chained_assignment = None  # default='warn'
        # calculate new column (voltage)
        self[col] = self[potential_0].to_numpy() - self[potential_1].to_numpy()
        pd.options.mode.chained_assignment = "warn"
        return self

    def get_all_voltages(self):
        """returns all voltages in the data frame.

        Returns
        -------
        list
            List containing all voltages present in self.columns as specifiers.
        """
        # iterate over columns and find voltages
        voltages = []
        for column in self.columns:  # pylint: disable=not-an-iterable
            try:
                if (column.specifier == specifiers.VOLTAGE) and (len(column.nodes) == 2):
                    voltages.append(column)
            except AttributeError:
                pass

        return voltages

    def get_all_potentials(self):
        """
        Returns
        -------
        list
            List containing all voltages present in self.columns as specifiers.
        """
        # iterate over columns and find voltages
        potentials = []
        for column in self.columns:  # pylint: disable=not-an-iterable
            try:
                if (column.specifier == specifiers.VOLTAGE) and (len(column.nodes) == 1):
                    potentials.append(column)
            except AttributeError:
                pass

        return potentials

    def get_all_nodes(self):
        """Gives a set of all nodes in voltages of a given df.

        Returns
        -------
        set
            Set containing all nodes of a given device.
        """
        # iterate over columns and find nodes
        nodes = set()
        for voltage in self.columns:  # pylint: disable=not-an-iterable
            if voltage.specifier == specifiers.VOLTAGE:
                if voltage.sub_specifiers == []:
                    nodes.update(set(voltage.nodes))
                else:
                    nodes.update(
                        set([node + str(voltage.sub_specifiers) for node in voltage.nodes])
                    )

        return nodes

    def create_potential(self, nodes, reference_node, voltage_sub_specifiers=None, debug=False):
        """Try to create col from existing data in dataframe df.

        Parameters
        ----------
        nodes  :  [node]
            List containing the nodes present in self or just directly list of nodes.
        reference_node : str
            Name of the reference node.
        voltage_sub_specifiers : [str]
            The sub specifier that shall be valid for the voltage to be created by this method.
        debug : bool, False
            If True, print out debugging information

        Returns
        -------
        df  : :class:`DMT.core.DataFrame`
            DMT.Dataframe with new dataframe, if successfull

        Raises
        ------
        IOError
            If it is not possible to create the potential
        """
        # init specifiers for the potentials
        potential = SpecifierStr(
            specifiers.VOLTAGE, nodes[0], sub_specifiers=voltage_sub_specifiers
        )

        try:
            reference_potential = SpecifierStr(
                specifiers.VOLTAGE, reference_node, sub_specifiers=voltage_sub_specifiers
            )
        except TypeError as err:
            raise IOError("DMT->data_frame: No or incompatible reference_node were given.") from err

        if reference_potential not in self.columns:  # pylint: disable=unsupported-membership-test
            potentials = self.get_all_potentials()
            # try create potential by faking different reference potential
            for potential in potentials:
                try:
                    self.create_potential(
                        [reference_node],
                        potential.nodes[0],
                        voltage_sub_specifiers=voltage_sub_specifiers,
                    )
                except IOError as err:
                    if (
                        "DMT->data_frame: Potentials and voltages are uncoupled -> can not create potential."
                        in err.args
                    ):
                        continue
                    else:
                        raise

            # is it set now?
            if reference_potential in self.columns:  # pylint: disable=unsupported-membership-test
                # it is set now -> shift all potentials so that the reference is 0
                if any(self[reference_potential]):
                    for potential_to_shift in self.get_all_potentials():
                        if (
                            potential_to_shift != reference_potential
                        ):  # this is the last one to shift...
                            self[potential_to_shift] = (
                                self[potential_to_shift] - self[reference_potential]
                            )

                    # now all potentials are relativ to the new reference....
                    self[reference_potential] = 0
            else:
                # all other potentials are independent of the reference or no potential is set -> create new reference as ground
                self[reference_potential] = 0

            # it can happen, that the potential is calculated as intermediate result...
            if potential in self.columns:  # pylint: disable=unsupported-membership-test
                return self

        if nodes[0] == reference_node:
            return self

        # is there a connection between all potentials and all voltages?
        voltages = self.get_all_voltages()
        nodes_in_potentials = [pot.nodes[0] for pot in self.get_all_potentials()]
        nodes_in_voltages = flatten([volt.nodes for volt in voltages])
        if not any(
            [node_in_voltages in nodes_in_potentials for node_in_voltages in nodes_in_voltages]
        ):
            raise IOError(
                "DMT->data_frame: Potentials and voltages are uncoupled -> can not create potential."
            )

        # is there a direct voltage?
        for voltage in voltages:
            potentials_in_voltage = [
                SpecifierStr(specifiers.VOLTAGE, node, sub_specifiers=voltage.sub_specifiers)
                for node in voltage.nodes
            ]
            if potential in potentials_in_voltage:
                for other_potential in [
                    SpecifierStr(specifiers.VOLTAGE, node, sub_specifiers=voltage.sub_specifiers)
                    for node in voltage.nodes
                    if node != nodes[0]
                ]:
                    # try to find a direct voltage from a other potential first
                    if (
                        other_potential in self.columns
                    ):  # pylint: disable=unsupported-membership-test
                        if nodes[0] == potentials_in_voltage[0].nodes[0]:
                            if debug:
                                print(
                                    "during setting of potential " + nodes[0] + " -> set potential"
                                )
                                print(
                                    str(potential)
                                    + " = "
                                    + str(voltage)
                                    + "+"
                                    + str(other_potential)
                                )
                            self[potential] = self[voltage] + self[other_potential]
                            break
                        else:
                            if debug:
                                print(
                                    "during setting of potential " + nodes[0] + " -> set potential"
                                )
                                print(
                                    str(potential)
                                    + " = "
                                    + str(other_potential)
                                    + "-"
                                    + str(voltage)
                                )
                            self[potential] = self[other_potential] - self[voltage]
                            break

        # did we reach our goal ?
        if potential in self.columns:  # pylint: disable=unsupported-membership-test
            return self

        # create other intermediate potential first and try again
        for voltage in voltages:
            potentials_in_voltage = [
                SpecifierStr(specifiers.VOLTAGE, node, sub_specifiers=voltage.sub_specifiers)
                for node in voltage.nodes
            ]
            if potential in potentials_in_voltage:
                other_potential = next(
                    SpecifierStr(specifiers.VOLTAGE, node, sub_specifiers=voltage.sub_specifiers)
                    for node in voltage.nodes
                    if node != nodes[0]
                )
                if (
                    other_potential not in self.columns
                ):  # pylint: disable=unsupported-membership-test
                    if debug:
                        print(
                            "during setting of potential "
                            + nodes[0]
                            + " -> will create "
                            + str(other_potential)
                        )
                    self.create_potential(
                        other_potential.nodes,
                        reference_node,
                        voltage_sub_specifiers=voltage_sub_specifiers,
                        debug=debug,
                    )

                    if nodes[0] == potentials_in_voltage[0].nodes[0]:
                        if debug:
                            print("during setting of potential " + nodes[0] + " -> set potential")
                            print(
                                str(potential) + " = " + str(voltage) + "+" + str(other_potential)
                            )
                        self[potential] = self[voltage] + self[other_potential]
                        break
                    else:
                        if debug:
                            print("during setting of potential " + nodes[0] + " -> set potential")
                            print(
                                str(potential) + " = " + str(other_potential) + "-" + str(voltage)
                            )
                        self[potential] = self[other_potential] - self[voltage]
                        break

        return self

    def drop_all_voltages(self):
        """Drops all found voltages (but ignores potentials)

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DMT.Dataframe with new dataframe, if successfull
        """
        to_drop = []
        # pylint: disable = not-an-iterable
        for col in self.columns:
            try:
                if col.specifier == specifiers.VOLTAGE and len(col.nodes) == 2:
                    to_drop.append(col)
            except AttributeError:  # if it is not a specifier we cant do anything
                pass

        return self.drop(labels=to_drop, axis=1)

    def parallel_norm(self, n_parallel, port_1, port_2):
        """Normalize the data in df with regard to the number of devices in parallel.

        This method normalizes S-parameters and currents with regard to the number
        of devices of the same geometry connected in parallel.

        Parameters
        ----------
        n_parallel    : int
            number of parallel devices
        port_1        : str
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : str
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DataFrame containing the normalized S parameters and currents.
        """
        # normalization of S parameters and deletion of all other parameters
        try:
            s_para_values = self.get_ss_para("S", port_1, port_2)
            s_para_values = DataFrame.processor.parallel_norm(s_para_values, n_parallel)
            self = self.set_ss_para("S", s_para_values, port_1, port_2)
        except StopIteration:
            pass

        # pylint: disable = not-an-iterable
        for col in self.columns:
            if specifiers.CURRENT in col:
                self[col] = self[col] / n_parallel

        return self

    def deembed_short(self, df_short, ports, times=1):
        """Deembed the measured data in df from the measured data in df_short.

        This method deembeds the measured small signal parameters in df using the measured small signal parameters in df_short.

        Parameters
        ----------
        df_short    :  :class:`DMT.core.DataFrame`
            df containing the S parameters of a short dummy.
        ports       :  [str]
            List of port names. Should be the same for the device and the open and short.  Usually this would be ['1', '2']. For a BJT in common emitter, should be ['B', 'C', ...]
        times       : int
            The number of times the short has to be removed.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DataFrame containing the de-embedded S parameters.
        """
        s_para_values = self.get_ss_para("S", *ports)
        s_para_short_values = df_short.get_ss_para("S", *ports)

        freq_short = df_short[specifiers.FREQUENCY]
        freq_self = self[specifiers.FREQUENCY]
        freq_self_unique = np.unique(freq_self)
        freq_short_unique = np.unique(freq_short)

        if np.all(freq_self_unique == freq_short_unique):  # frequencies match exactly
            s_para_values = DataFrame.processor.deembed_short(
                s_para_values, s_para_short_values, times=times
            )
        elif len(freq_short_unique) > len(
            freq_self_unique
        ):  # try to throw away frequencies in short
            indices_to_delete = [
                i for i, _freq in enumerate(freq_short_unique) if not _freq in freq_self_unique
            ]
            s_para_short_values = np.delete(s_para_short_values, indices_to_delete, 0)
            s_para_values = DataFrame.processor.deembed_short(
                s_para_values, s_para_short_values, times=times
            )

        # set the de-embedded parameters and remove un-necessary small signal parameters
        self = self.set_ss_para("S", s_para_values, *ports)
        self = self.strip_ss_para()
        return self

    def deembed_open(self, df_open, ports, times=1):
        """Deembed the measured data in df from the measured data in df_open.

        This method deembeds the masured small signal parameters in df using the measured small signal parameters in df_open.

        Parameters
        ----------
        df_open     :  :class:`DMT.core.DataFrame`
            df containgin the measured open structure.

        ports       :  [str]
            List of port names. Should be the same for the device and the open and short.
            Usually this would be ['1', '2']. For a BJT in common emitter, should be ['B', 'C', ...]
        times       :  int
            Numer of times the open has to be removed.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DataFrame containing the de-embedded S parameters.

        """
        try:
            s_para_values = self.get_ss_para("S", *ports)
            s_para_open_values = df_open.get_ss_para("S", *ports)

            freq_open = df_open[specifiers.FREQUENCY]
            freq_self = self[specifiers.FREQUENCY]
            freq_self_unique = np.unique(freq_self)
            freq_open_unique = np.unique(freq_open)

            if np.all(freq_self_unique == freq_open_unique):  # frequencies match exactly
                s_para_values = DataFrame.processor.deembed_open(
                    s_para_values, s_para_open_values, times=times
                )
            elif len(freq_open_unique) > len(
                freq_self_unique
            ):  # try to throw away frequencies in open
                indices_to_delete = [
                    i for i, _freq in enumerate(freq_open_unique) if not _freq in freq_self_unique
                ]
                s_para_open_values = np.delete(s_para_open_values, indices_to_delete, 0)
                s_para_values = DataFrame.processor.deembed_open(
                    s_para_values, s_para_open_values, times=times
                )

            else:
                raise IOError(
                    "DMT.core.DataFrame:Something went wrong during deembeeding. Maybe frequencies do not match."
                )
        except ValueError:
            raise ValueError

        # set values and throw away unnecessary parameters
        self = self.set_ss_para("S", s_para_values, *ports)
        self = self.strip_ss_para()
        return self

    def deembed(self, df_open, df_short, ports=None, ndevices=1, ndevices_open=1, ndevices_short=1):
        """Deembed the measured data in df from the measured data in df_open and df_short.

        This method deembeds the masured small signal parameters in df using the measured small signal parameters of one dummy open and one dummy short structure.

        Parameters
        ----------
        df_open     :  :class:`DMT.core.DataFrame`
            df containgin the measured open structure.

        df_short    :  :class:`DMT.core.DataFrame`
            df containing the measured short structure.

        ports       :  [str]
            List of port names. Should be the same for the device and the open and short.
            Defaults to ['1', '2']. For a BJT in common emitter, should be ['B', 'C', ...]

        ndevices : int
            Number of parallel devices to be deembedded.
        ndevices_open : int
            Number of parallel devices of the open structure.
        ndevices_short : int
            Number of parallel devices of the short structure.

        Returns
        -------
          :class:`DMT.core.DataFrame`
            DataFrame containing the de-embedded S parameters.

        Notes
        -----
        ..todo: make this method remember df_open and df_short.
        """
        # pr = cProfile.Profile()
        # pr.enable()
        if ports is None:
            ports = ["1", "2"]

        # short and open only onces de-embedded, this makes no sense:
        # if ndevices_open != ndevices_short:
        #     raise IOError(
        #         "DMT -> DataFrame -> deembed: The number of open and short devices must be exactly equal."
        #     )
        # if ndevices < ndevices_short:
        #     raise IOError(
        #         "DMT -> DataFrame -> deembed: The number of parallel devices is smaller then the number of shorts. Not implemented."
        #     )

        # find out how many "times" the deembedding structure needs to be deembedded
        times = ndevices / ndevices_short
        times = float(1)
        if not times.is_integer():
            raise IOError(
                "DMT -> DataFrame -> deembed: The number of devices divided by the number of parallel deembedding structures is not an integer."
            )

        try:
            # create a new object, else the df_short will deembed itself with every call of this method!
            df_short_deem = copy.deepcopy(df_short)
            self = self.deembed_open(df_open, ports, times=times)
            df_short_deem = df_short_deem.deembed_open(df_open, ports)
            self = self.deembed_short(df_short_deem, ports, times=times)
        except ValueError:
            raise ValueError
        # pr.disable()
        # pr.print_stats(sort='cumtime')
        return self

    def determine_mres(self, forced_current=False):
        """Determine the external restistances caused by the metallization.

        Parameters
        ----------
        forced_current : bool, optional


        Returns
        -------
        dict
            A dict of resistances Rb,m, Rc,m, and Re,m.

        """

        if forced_current:

            try:

                df_RC = self.loc[np.isclose(self[specifiers.CURRENT + "B"], 0.0, atol=1e-6)]
                df_RB = self.loc[np.isclose(self[specifiers.CURRENT + "C"], 0.0, atol=1e-6)]
            except KeyError:
                raise IOError

            mres = {}

            re2 = np.polyfit(df_RB["I_B"], df_RB["V_C"], 1)[0]
            re1 = np.polyfit(df_RC["I_C"], df_RC["V_B"], 1)[0]

            re = (re1 + re2) / 2
            de = np.abs(re1 - re2)

            print(
                f"rem disagreement is {de} Ohm. Using average {re} Ohm with uncertantiy {de / (200 * re)}%"
            )

            rc = np.polyfit(df_RC["I_C"], df_RC["V_C"], 1)[0] - re
            rb = np.polyfit(df_RB["I_B"], df_RB["V_B"], 1)[0] - re

            return {"R_CM": rc, "R_BM": rb, "R_EM": re}
        else:
            try:
                df_RCE_RBE = self.loc[
                    np.isclose(
                        self[specifiers.VOLTAGE + "C"], self[specifiers.VOLTAGE + "E"], atol=1e-4
                    )
                ]
                df_RBC = self.loc[
                    np.isclose(
                        self[specifiers.VOLTAGE + "B"], self[specifiers.VOLTAGE + "E"], atol=1e-4
                    )
                ]
            except KeyError:
                raise IOError

            mres = {}

            mres = DataFrame.processor.calc_RBC_RBE(mres, df_RCE_RBE)
            mres = DataFrame.processor.calc_RCE(mres, df_RBC)

            mres = DataFrame.processor.convert_mres(mres)

            return mres

    def deembed_DC(self, mres=None, df_short_dc=None, forced_current=False):
        """Deembed the measured DC data in df from external metallization resistances.

        Determine the metallization resistances and substract their impact from the measured voltages.

        Parameters
        ----------
        mres : dict
        df_short_dc : :class:`DMT.core.DataFrame`
            df containing the dc-measured short structure.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DataFrame containing the de-embedded DC-voltages.

        """

        if mres is None:
            mres = df_short_dc.determine_mres(forced_current=forced_current)

        return DataFrame.processor.deembed_mres(self, mres)

    def check_ss_cols(self, para):
        """Check the existence of the small signal parameters para cols.

        Parameters
        ----------
        para : string
            Small signal parameter whose existence shall be checked.

        Raises
        ------
        IOError
            Raised if the ss parameters para are not present in self.columns.

        """
        for col in self.columns:  # pylint: disable=not-an-iterable
            try:
                if col.specifier == para:
                    return True
            except AttributeError:
                if col[0:2] == para + "_":  # we should no allow this :(
                    return True

        return False

    def get_ss_para(self, para, port_1, *ports_n):
        """Return an array of the small signal (ss) parameters para in df.

        Parameters
        ----------
        para            :  string
            Name of the parameter to be returned
        port_1          : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        *port_n         : string
            Further nodes in order to build up a higher order matrix.

        Returns
        -------
        np.ndarray()
            Numpy array containing the values of the ss parameter para with shape [n_freq,n_port_in,n_port_out].
        """
        # check existence of the paras
        para_desired = para
        try:
            para_avail = [
                ss_para
                for ss_para in [para] + [speci for speci in specifiers_ss_para]
                if self.check_ss_cols(ss_para)
            ]
            if len(para_avail) == 0:
                raise StopIteration
            elif len(para_avail) == 1:
                para_avail = para_avail[0]
            else:
                if para_desired in para_avail:
                    para_avail = para_desired
                else:
                    para_avail = para_avail[0]
        except StopIteration:
            logging.warning("Warning: DMT->DataFrame: No SS-Parameters were found in DataFrame.")
            raise StopIteration

        list_ports = [port_1]
        columns = self.columns
        for node in ports_n:
            # check if full set of this parameter is here for this node
            col_0 = SpecifierStr(para_avail.upper(), port_1, node)
            col_1 = SpecifierStr(para_avail.upper(), node, port_1)
            col_2 = SpecifierStr(para_avail.upper(), node, node)

            # did not check for node[i], node[j] on purpose, because node[j] may be not part of the full set..
            if (
                col_0 in columns and col_1 in columns and col_2 in columns
            ):  # pylint: disable=unsupported-membership-test
                list_ports.append(node)

        # get the existing parameters from df and put them into a numpy array
        try:
            n_freq = len(self[specifiers.FREQUENCY])
        except KeyError:
            raise IOError("DMT -> data_frame -> get_ss_para: no column FREQ in data frame.")

        n_port = len(list_ports)

        para_values = np.zeros((n_freq, n_port, n_port), dtype=np.complex128)
        for i in range(n_port):
            for j in range(n_port):
                para_values[:, i, j] = self.loc[
                    :, SpecifierStr(para_avail.upper(), list_ports[i], list_ports[j])
                ]

        # we found some parameters, however not the desired ones. Try converting
        if para_avail != para_desired:
            para_values = self.processor.convert_n_port_para(
                para_values, p_from=str(para_avail), p_to=str(para_desired)
            )

        return para_values

    def set_ss_para(self, para, para_values, port_1, *ports_n):
        """Create or update the columns corresponding to small signal (ss) parameters para in df using the values in para_values.

        Parameters
        ----------
        para          :  string
            Name of the parameter to be set.
        para_values   :  np.array()
            Array that holds the values of the ss parameter with shape [n_freq,n_port,n_port].
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        *port_n       : string
            Further nodes in order to build up a higher order matrix.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            DataFrame that contains the columns for the ss parameter para with values according to para_values.

        """
        n_port = para_values.shape[1]
        list_nodes = [port_1] + list(ports_n)
        for i in range(n_port):
            for j in range(n_port):
                para_str = SpecifierStr(para, list_nodes[i], list_nodes[j])
                self[para_str] = para_values[:, i, j]

        return self

    def strip_ss_para(self, keep="S"):
        """Throw away all but keep small signal parameters.

        Parameters
        ----------
        keep  :  string
            Default='S'. SS Para that shall be kept.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            The initial DataFrame without all ss parameters except keep.
        """
        ss_paras = self.get_all_ss_para()
        drop = []
        for para in ss_paras:
            if para.specifier != keep:
                drop.append(para)

        return self.drop(columns=drop)

    def get_all_ss_para(self):
        """Return a list of strings of all small signal parameters (ss) in df.

        Returns
        -------
        [str]
            List of strings of the ss parameter columns in self.
        """
        ss_paras = []
        for col in self.columns:  # pylint: disable=not-an-iterable
            try:
                if col.specifier in specifiers_ss_para:
                    ss_paras.append(col)
            except AttributeError:
                pass

        return ss_paras

    def calc_ft(self, port_1, port_2):
        """Calculates the transit frequency FT using the spot frequency method.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains F_T.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self.loc[:, specifiers.TRANSIT_FREQUENCY] = self.processor.calc_ft(
            self.loc[:, specifiers.FREQUENCY], s_para_values, "S"
        )
        return self

    def calc_tfit2(self, port_1, port_2):
        """Calculates the transit frequency FT using the spot frequency method.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains F_T.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self.loc[:, specifiers.TRANSIT_TIME] = self.processor.calc_tfit2(
            self[specifiers.FREQUENCY], s_para_values, "S"
        )
        return self

    def calc_tfit(self, tfit_kind: int, ports):
        port1 = ports[0]
        port2 = ports[1]
        if tfit_kind == 1:
            return self.calc_tfit1(port1, port2)
        elif tfit_kind == 2:
            return self.calc_tfit2(port1, port2)
        elif tfit_kind == 3:
            self.ensure_specifier_column(specifiers.TRANSIT_TIME, ports=ports)
            return self

    def calc_tfit1(self, port_1, port_2):
        """Calculates the transit frequency FT using the spot frequency method.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains F_T.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self.loc[:, specifiers.TRANSIT_TIME] = self.processor.calc_tfit1(
            self[specifiers.FREQUENCY], s_para_values, "S"
        )
        return self

    def calc_fmax(self, port_1, port_2):
        """Calculates the maximum frequency of oscillation FMAX.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains F_MAX.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self[specifiers.MAXIMUM_OSCILLATION_FREQUENCY] = self.processor.calc_fmax(
            self[specifiers.FREQUENCY], s_para_values, "S"
        )
        return self

    def calc_msg(self, port_1, port_2):
        """Calculates the maximum stable gain MSG.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains MSG.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self["MSG"] = self.processor.calc_msg(self["FREQ"], s_para_values, "S")
        return self

    def calc_unilateral_gain(self, port_1, port_2):
        """Calculates the unilateral gain `GU <https://www2.eecs.berkeley.edu/Pubs/TechRpts/2016/EECS-2016-15.pdf>`_ .

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
         :class:`DMT.core.DataFrame`
            Dataframe that contains GU.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self["GU"] = self.processor.calc_unilateral_gain(self["FREQ"], s_para_values, "S")
        return self

    def calc_mag(self, port_1, port_2):
        """Calculates the maximum available gain MAG.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains MAG.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self["MAG"] = self.processor.calc_mag(self["FREQ"], s_para_values, "S")
        return self

    def calc_k(self, port_1, port_2):
        """Calculates the k-factor K.

        Parameters
        ----------
        port_1        : string
            Name of the nodes of the port 1, for a BJT in common-emitter this is 'B'.
        port_2        : string
            Name of the nodes of the port 2, for a BJT in common-emitter this is 'C'.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains K.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        self["K"] = self.processor.calc_k(self["FREQ"], s_para_values, "S")
        return self

    def calc_cbe(self, port_1="B", port_2="C"):
        """Calculates the base-emitter junction capacitance CBE assuming PI equivalent circuit and common emitter configuration.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains CBE.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        sp_cbe = specifiers.CAPACITANCE + ["B", "E"]

        # put values in col of self
        pd.options.mode.chained_assignment = (
            None  # default='warn' , Markus: This should not warn here.
        )
        if port_1 == "B" and port_2 == "C":
            self[sp_cbe] = self.processor.calc_cap_shunt_port_1(self["FREQ"], s_para_values, "S")
        elif port_1 == "C" and port_2 == "B":
            self[sp_cbe] = self.processor.calc_cap_shunt_port_2(self["FREQ"], s_para_values, "S")
        elif port_1 == "B" and port_2 == "E":
            self[sp_cbe] = self.processor.calc_cap_series_thru(self["FREQ"], s_para_values, "S")
        elif port_1 == "E" and port_2 == "B":
            self[sp_cbe] = self.processor.calc_cap_series_thru(self["FREQ"], s_para_values, "S")
        elif port_1 == "E" and port_2 == "C":  # common base
            self[sp_cbe] = self.processor.calc_cap_shunt_port_1(self["FREQ"], s_para_values, "S")
        elif port_1 == "E" and port_2 == "C":  # common base
            self[sp_cbe] = self.processor.calc_cap_shunt_port_2(self["FREQ"], s_para_values, "S")
        elif port_1 == "C" and port_2 == "E":  # common base
            self[sp_cbe] = self.processor.calc_cap_shunt_port_1(self["FREQ"], s_para_values, "S")
        else:
            raise NotImplementedError(
                "DMT -> DataFrame -> calc_cbe: transistor configuration not implemented."
            )

        pd.options.mode.chained_assignment = "warn"
        return self

    def calc_cce(self, port_1="B", port_2="C"):
        """Calculates the collector-emitter junction capacitance CCE, assuming PI equivalent circuit and common emitter configuration with base at port 1.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains CCE.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        # put values in col of self
        pd.options.mode.chained_assignment = (
            None  # default='warn' , Markus: This should not warn here.
        )
        self[specifiers.CAPACITANCE + ["C", "E"]] = self.processor.calc_cap_shunt_port_2(
            self["FREQ"], s_para_values, "S"
        )
        pd.options.mode.chained_assignment = "warn"
        return self

    def calc_cbc(self, port_1="B", port_2="C"):
        """Calculates the base-collector junction capacitance CBC.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains CBE.
        """
        # get values
        s_para_values = self.get_ss_para("S", port_1, port_2)

        sp_cbc = specifiers.CAPACITANCE + ["B", "C"]

        # put values in col of self
        if port_1 == "B" and port_2 == "C":
            self[sp_cbc] = self.processor.calc_cap_series_thru(self["FREQ"], s_para_values, "S")
        elif port_1 == "C" and port_2 == "B":
            self[sp_cbc] = self.processor.calc_cap_series_thru(self["FREQ"], s_para_values, "S")
        elif port_1 == "B" and port_2 == "E":
            self[sp_cbc] = self.processor.calc_cap_shunt_port_1(self["FREQ"], s_para_values, "S")
        elif port_1 == "E" and port_2 == "B":
            self[sp_cbc] = self.processor.calc_cap_shunt_port_2(self["FREQ"], s_para_values, "S")
        elif port_1 == "C" and port_2 == "E":
            self[sp_cbc] = self.processor.calc_cap_shunt_port_1(self["FREQ"], s_para_values, "S")
        elif port_1 == "E" and port_2 == "C":
            self[sp_cbc] = self.processor.calc_cap_shunt_port_2(self["FREQ"], s_para_values, "S")
        else:
            raise NotImplementedError(
                "DMT -> DataFrame -> calc_cbe: transistor configuration not implemented."
            )
        return self

    def calc_beta(self):
        """Calculates the DC current gain of a bipolar transistor

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains the DC current amplficiation I_C/I_B
        """
        # put values in col of self
        col_ic = specifiers.CURRENT + "C"
        col_ib = specifiers.CURRENT + "B"
        self[specifiers.DC_CURRENT_AMPLIFICATION] = self.processor.calc_beta(
            self[col_ic], self[col_ib]
        )
        return self

    def calc_gm(self):
        """Calculates the DC transconductance of a BJT.

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains the TRANSCONDUCTANCE
        """
        col_ic = specifiers.CURRENT + "C"
        col_vc_forced = specifiers.VOLTAGE + "C" + sub_specifiers.FORCED
        col_vb_forced = specifiers.VOLTAGE + "B" + sub_specifiers.FORCED
        col_vbe = specifiers.VOLTAGE + ["B", "E"]

        # get voltages
        self.ensure_specifier_column(col_vbe)
        self.ensure_specifier_column(col_ic)

        # if possible also pass vbc forced
        try:
            self.ensure_specifier_column(col_vc_forced)
            self.ensure_specifier_column(col_vb_forced)
            try:
                self.loc[:, specifiers.TRANSCONDUCTANCE] = self.processor.calc_gm(
                    self[col_ic].to_numpy(),
                    self[col_vbe].to_numpy(),
                    vc_forced=self[col_vc_forced].to_numpy(),
                    vb_forced=self[col_vb_forced].to_numpy(),
                )
            except ValueError:
                pass
        except KeyError:
            self.loc[:, specifiers.TRANSCONDUCTANCE] = self.processor.calc_gm(
                self[col_ic].to_numpy(), self[col_vbe].to_numpy()
            )

        return self

    def smooth_SS_para(self, ports, fmin):
        """Smooth the Small signal parameters in self and return a new dataframe.

        Method from:
        B. Saha et al.,
        "Reliable Technology Evaluation of SiGe HBTs and MOSFETs: fMAX Estimation From Measured Data,"
        in IEEE Electron Device Letters, vol. 42, no. 1, pp. 14-17, Jan. 2021, doi: 10.1109/LED.2020.3040891.
        to calculate fmax
        It is assumed that the frequency is strictly increasing for every DC operating point in the DataFrame,
        as is usually the case for measurement data-

        Parameters
        ----------
        ports : [str]
            The AC ports of the device
        fmin : float
            Minimum frequency to consider for smoothing

        Returns
        -------
        :class:`DMT.core.DataFrame`
            Dataframe that contains the smoothed Y Parameters
        """
        # make sure we have all Y parameters available
        for port1 in ports:
            for port2 in ports:
                for kind in [sub_specifiers.REAL, sub_specifiers.IMAG]:
                    self.ensure_specifier_column(
                        specifiers_ss_para.SS_PARA_Y + port1 + port2 + kind,
                        ports=["B", "C"],
                    )

        # assume this is real measurement data: freq is strictly increasing and we can just always get next N_FREQ values
        freq = np.unique(self[specifiers.FREQUENCY].to_numpy())
        n_freq = len(freq)
        n_rows = len(self.index)
        n_slices = int(n_rows / n_freq)

        df_new = copy.deepcopy(self)
        # remove all small signal parameters in df_new and prepare new columns
        df_new = df_new.strip_ss_para(keep="S")
        df_new = df_new.strip_ss_para(keep="Y")
        for port1 in ports:
            for port2 in ports:
                for kind in [sub_specifiers.REAL, sub_specifiers.IMAG]:
                    y_para = specifiers_ss_para.SS_PARA_Y + port1 + port2 + kind
                    df_new[y_para] = 0

        # actual smoothing
        for port1 in ports:
            for port2 in ports:
                for kind in [sub_specifiers.REAL, sub_specifiers.IMAG]:
                    y_para = specifiers_ss_para.SS_PARA_Y + port1 + port2 + kind
                    for i in range(n_slices):
                        i_low = n_freq * i
                        i_upp = n_freq * (i + 1)

                        freq_i = self.iloc[i_low:i_upp][specifiers.FREQUENCY].to_numpy()
                        indices = freq_i > fmin

                        y_raw = self.iloc[i_low:i_upp][y_para].to_numpy()

                        freq_i_lim = self.iloc[i_low:i_upp][specifiers.FREQUENCY].to_numpy()[
                            indices
                        ]
                        y_raw_lim = y_raw[indices]

                        if kind == sub_specifiers.REAL:
                            fun = fun_re
                            norm = 1
                            norm_lim = 1
                        else:
                            fun = fun_im
                            norm_lim = 2 * np.pi * freq_i_lim
                            norm = 2 * np.pi * freq_i

                        popt, _pcov = curve_fit(fun, np.log10(freq_i_lim), y_raw_lim / norm_lim)
                        y_fitted_lim = fun(np.log10(freq_i_lim), *popt)

                        df_new.iloc[i_low:i_upp, df_new.columns.get_loc(y_para)] = (
                            fun(np.log10(freq_i), *popt) * norm
                        )

        # merge real and complex part, remove afterwards
        for port1 in ["B", "C"]:
            for port2 in ["B", "C"]:
                y_para = specifiers_ss_para.SS_PARA_Y + port1 + port2
                y_para_re = specifiers_ss_para.SS_PARA_Y + port1 + port2 + sub_specifiers.REAL
                y_para_im = specifiers_ss_para.SS_PARA_Y + port1 + port2 + sub_specifiers.IMAG
                df_new[y_para] = df_new[y_para_re] + 1j * df_new[y_para_im]
                df_new.drop(columns=[y_para_im, y_para_re])

        return df_new

    def iter_unique_col(self, column, decimals=5):
        """Allows iteration over the unique values and their slices of a column

        Parameters
        ----------
        column : :class:`DMT.core.SpecifierStr` or str
            Column name to iterate over
        decimals : int, optional
            Rounding precision for the unique, None to turn off, by default 5

        Returns
        -------
        :class:`DMT.core.data_frame.IterUniqueRoundColumn`
            Iterator over all unique values of this column

        Examples
        --------
        We want to replace:

        >>> for vbc in np.unique(np.round(df[col_vbc], decimals=5)):
                data = df[np.isclose(df[col_vbc], atol=1e-5)]
                ... # user action

        with

        >>> for index, vbc, data in df.iter_unique_col(col_vbc, decimals=5):
                ... # user action

        """
        return IterUniqueRoundColumn(self, column, decimals=decimals)


class IterUniqueRoundColumn(object):
    """Iterator to iterate over rounded unique values of a column

    Parameters
    ----------
    dataframe : :class:`DMT.core.Dataframe`
        DMT (or parandas) dataframe with data including the column to iterate over.
    column : :class:`DMT.core.SpecifierStr` or str
        Column name to iterate over
    decimals : int, optional
        Rounding precision for the unique, None to turn off, by default 5

    """

    def __init__(self, dataframe, column, decimals=5):
        self.index = 0  # start at 0
        self.dataframe = dataframe
        self.column = column

        if decimals is None:
            self.val_unique = np.unique(dataframe[column])
            self.atol = None
        else:
            self.val_unique = np.unique(np.round(dataframe[column], decimals=decimals))
            self.atol = 5 * np.float_power(10, -(decimals + 1))

    def __len__(self):
        """Convenience to so len(iterator) can be used.

        Returns
        -------
        int
            Length of the uniqued values
        """
        return len(self.val_unique)

    def __iter__(self):
        """Here self is returned -> this class itself is an iterator.

        Returns
        -------
        :class:`DMT.core.data_frame.IterUniqueRoundColumn`
            Iterator over all unique values of this column
        """
        return self

    def __next__(self):
        """This routine is magic and sets the iteration behavior.

        Returns
        -------
        int
            Index of current iteration
        float
            Value of the uniqued column
        :class:`DMT.core.Dataframe`
            Slice of the dataframe with the value

        Raises
        ------
        StopIteration
            As soon as all values are iterated
        """
        if self.index == len(self):
            # end reached
            raise StopIteration

        val = self.val_unique[self.index]

        if self.atol is None:
            dataframe = self.dataframe[self.dataframe[self.column] == val]
        else:
            dataframe = self.dataframe[np.isclose(self.dataframe[self.column], val, atol=self.atol)]

        self.index += 1
        return self.index - 1, val, dataframe



"""testing the plotting data delivery. Plotting is not possible in the CI...
"""
import os
from pathlib import Path
from DMT.core import read_data, Plot, specifiers

folder_path = Path(__file__).resolve().parent


def read_mdm():
    df = read_data(folder_path / "test_data" / "short_freq.mdm")
    df = read_data(folder_path / "test_data" / "short_dc.mdm")


def test_label_generation():
    plt_test = Plot("y_1(x)")

    assert plt_test.x_label == "$x$"
    assert plt_test.y_label == "$y_1$"

    plt_test = Plot(
        "y_2(x)",
        x_label="x/m",
        y_label="y_2/A",
    )

    assert plt_test.x_label == "x/m"
    assert plt_test.y_label == "y_2/A"

    plt_test = Plot(
        "y_2(x)",
        x_specifier=specifiers.X,
        x_scale=1e9,
        y_specifier=specifiers.NET_DOPING,
        # y_scale=1e6, # use natural scale
        y_log=True,
    )

    assert plt_test.x_label == "$x_{\\mathrm{}}\\left(\\si{\\nano\\meter}\\right)$"
    assert plt_test.y_label == "$N_{\\mathrm{net}}\\left(\\si{\\per\\centi\\meter\\cubed}\\right)$"


# def test_plot_add_line():

if __name__ == "__main__":
    test_label_generation()



""" Tests all known cases for clean_names

"""
import os
import logging

from DMT.exceptions import UnknownColumnError
from DMT.core.dut_type import DutType
from DMT.core.naming import get_nodes

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=os.path.join("logs", "test_get_nodes.log"),
    filemode="w",
)


def test_get_nodes_teledyne():

    nodes_tlm = DutType.tlm.get_nodes()
    fallback_tlm = {"L": "L", "S": "R", "M": "M", "LI": "", "SI": "", "C": ""}
    # vli             vc              vm              vl              vs              il              ic              is
    cols_tlm = ["V_LI", "V_C", "V_M", "V_L", "V_S", "I_L", "I_C", "I_S"]
    cols_dmt_tlm = ["V_LI", "V_C", "V_M", "V_L", "V_R", "I_L", "I_C", "I_R"]
    call_get_nodes(cols_tlm, nodes_tlm, fallback_tlm, cols_dmt_tlm)

    nodes_tetrode = DutType.tetrode.get_nodes()
    fallback_tetrode = {"B1_": "B1", "B2_": "B2", "C_": "C", "E_": "E"}
    # ve              ib1_            vb1_            ib2_            vb2_            ic_             ie_
    cols_tetrode = ["V_E", "I_B1_", "V_B1_", "I_B2_", "V_B2_", "I_C_", "I_E_"]
    cols_dmt_tetrode = ["V_E", "I_B1", "V_B1", "I_B2", "V_B2", "I_C", "I_E"]
    call_get_nodes(cols_tetrode, nodes_tetrode, fallback_tetrode, cols_dmt_tetrode)

    nodes_bjt = DutType.bjt.get_nodes()
    fallback_bjt = {}
    # vb              vc              ic              ib
    # ICCAP_VAR vs         0              ICCAP_VAR ve         0              ICCAP_VAR vc         0.35           ICCAP_VAR vb         0.6
    cols_bjt = ["V_B", "V_C", "I_C", "I_B", "V_S", "V_E", "V_C", "V_B"]
    cols_dmt_bjt = ["V_B", "V_C", "I_C", "I_B", "V_S", "V_E", "V_C", "V_B"]
    call_get_nodes(cols_bjt, nodes_bjt, fallback_bjt, cols_dmt_bjt)

    # oder : ICCAP_VAR ve         0              ICCAP_VAR vs         0              ICCAP_VAR vbe        0.8            ICCAP_VAR vc         0.8
    cols_bjt = ["V_E", "V_S", "V_BE", "V_C"]
    cols_dmt_bjt = ["V_E", "V_S", "V_BE", "V_C"]
    call_get_nodes(cols_bjt, nodes_bjt, fallback_bjt, cols_dmt_bjt)


def test_get_nodes_DotSeven():

    nodes_tlm = DutType.tlm.get_nodes()
    fallback_tlm = {"S1": "L", "S3": "R", "S2": "M", "F": "", "FORCE": "L", "GND": "R"}
    # VF              VS1             VS2             VS3             IForce          IGnd
    # ICCAP_VAR IS2        0              ICCAP_VAR IS1        0              ICCAP_VAR IS3        0              ICCAP_VAR GND        0
    cols_tlm = ["V_F", "V_S1", "V_S2", "V_S3", "I_FORCE", "I_GND", "I_S2", "I_S1", "I_S3", "GND"]
    cols_dmt_tlm = ["V_F", "V_L", "V_M", "V_R", "I_L", "I_R", "I_M", "I_L", "I_R", "GND"]
    call_get_nodes(cols_tlm, nodes_tlm, fallback_tlm, cols_dmt_tlm)

    nodes_tetrode = DutType.tetrode.get_nodes()
    fallback_tetrode = {"BF1": "", "BF2": ""}
    # VE              VB1             VB2             IB1             IB2             IE              IC
    # ICCAP_VAR VBf1       -0.01          ICCAP_VAR VC         0              ICCAP_VAR VBf2       0.01
    cols_tetrode = ["V_E", "V_B1", "V_B2", "I_B1", "I_B2", "I_E", "I_C", "V_BF1", "V_C", "V_BF2"]
    cols_dmt_tetrode = [
        "V_E",
        "V_B1",
        "V_B2",
        "I_B1",
        "I_B2",
        "I_E",
        "I_C",
        "V_BF1",
        "V_C",
        "V_BF2",
    ]
    call_get_nodes(cols_tetrode, nodes_tetrode, fallback_tetrode, cols_dmt_tetrode)

    nodes_bjt = DutType.bjt.get_nodes()
    fallback_bjt = {}
    # vb              vc              ic              ib
    #  USER_VAR  vbc        0.5            ICCAP_VAR ve         0              ICCAP_VAR vs         0
    cols_bjt = ["V_B", "V_C", "I_C", "I_B", "V_BC", "V_E", "V_S"]
    cols_dmt_bjt = ["V_B", "V_C", "I_C", "I_B", "V_BC", "V_E", "V_S"]
    call_get_nodes(cols_bjt, nodes_bjt, fallback_bjt, cols_dmt_bjt)


def call_get_nodes(cols, nodes, fallback, cols_dmt):
    for i_col, col in enumerate(cols):
        if col[0] not in "VIC":
            # only rename the names of values which can be identified
            pass
        else:
            nodes_in_col = get_nodes(col, nodes, fallback=fallback)
            if nodes_in_col:
                col = col[0] + r"_" + "".join(nodes_in_col)
            else:
                raise UnknownColumnError(
                    "The column "
                    + col
                    + " is unknown to this DuT as the nodes can not be extracted! Either a fallback behavior or different DuT nodes for this column are needed."
                )

        assert col == cols_dmt[i_col]


if __name__ == "__main__":
    test_get_nodes_teledyne()
    test_get_nodes_DotSeven()
    dummy = 1



import setuptools
from itertools import chain

with open("README.md", "r") as fh:
    long_description = fh.read()

EXTRAS_REQUIRE = {
    "HDF5": ["tables"],
    "pyqtgraph": ["pyqtgraph"],
    "matplotlib": ["matplotlib"],
    "pyside2": ["PySide2"],
    "pyqt5": ["PyQt5"],
    "smithplot": ["matplotlib", "pysmithplot-3.10"],
    "develop": ["pylint", "black"],
    "latex": ["pylatex", "pylatexenc"],
    "remote": ["paramiko", "scp"],
}
EXTRAS_REQUIRE["full"] = list(set(chain(*EXTRAS_REQUIRE.values())))
EXTRAS_REQUIRE["full"].remove("PyQt5")  # not always needed

setuptools.setup(
    name="DMT_core",
    version="1.8.0-rc.1",
    author="M.Mueller, M.Krattenmacher",
    author_email="markus.mueller@semimod.de, mario.krattenmacher@semimod.de",
    description="Device Modeling Toolkit Core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/dmt-development/dmt",
    project_urls={
        "Bug Tracker": "https://gitlab.com/dmt-development/dmt",
        "Documentation": "https://dmt-development.gitlab.io/dmt-core/",
        "Source Code": "https://gitlab.com/dmt-development/dmt",
    },
    packages=setuptools.find_namespace_packages(include=["DMT.*"]),
    license="GNU GPLv3+",
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ],
    package_data={"": ["*.yaml", "*.txt", "*.tex", "*.bib"]},
    include_package_data=True,
    install_requires=[
        "scipy",
        "numpy",
        "scikit-rf",
        "reprint",
        "pandas",
        "joblib",
        "pytest",
        "pint",
        "pyyaml",
        "more_itertools",
        "colormath",
        "semver",
        "verilogae>=0.9b4",
        "h5py",
        "cycler",
        "colormath",
    ],
    extras_require=EXTRAS_REQUIRE,
)



from datetime import datetime
import numpy as np
from pathlib import Path
from DMT.core import read_data, DataFrame, specifiers


def test_data_processor_and_deem():
    # load test measurements
    df = read_data(Path(__file__).parent / "test_data" / "Spar_vb.mdm")
    df_open = read_data(Path(__file__).parent / "test_data" / "dummy_open_freq.mdm")
    df_short = read_data(Path(__file__).parent / "test_data" / "dummy_short_freq.mdm")

    # correct data format of measurements
    nodes = ["B", "C", "E", "S"]
    df = df.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_short = df_short.clean_data(nodes, "E", fallback={"1": "B", "2": "C"}, ac_ports=["B", "C"])
    df_open = df_open.clean_data(nodes, "E", fallback={"1": "B", "2": "C"}, ac_ports=["B", "C"])

    # deembed
    df = df.deembed_open(df_open, nodes)
    df_short = df_short.deembed_open(df_open, nodes)
    df = df.deembed_short(df_short, nodes)

    # calculate cbe
    df = df.calc_cbe()
    return df


def test_cap_extraction_common_base():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common base config PI network
    g1 = 1j * 2 * np.pi * freq * cbe
    g2 = 1j * 2 * np.pi * freq * cce
    g3 = 1j * 2 * np.pi * freq * cbc
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "E" + "E"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "C" + "C"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "C" + "E"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "E" + "C"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["E", "C"])  # slow
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["E", "C"])  # slow

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


def test_cap_extraction_common_base_reversed():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common base config PI network
    g3 = 1j * 2 * np.pi * freq * cbe
    g2 = 1j * 2 * np.pi * freq * cce
    g1 = 1j * 2 * np.pi * freq * cbc
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "E" + "E"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "C" + "C"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "C" + "E"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "E" + "C"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["C", "E"])  # slow
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["C", "E"])  # slow

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


def test_cap_extraction_common_emitter():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common emitter config PI network
    g1 = 1j * 2 * np.pi * freq * cbe
    g2 = 1j * 2 * np.pi * freq * cbc
    g3 = 0
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "B" + "B"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "C" + "C"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "B" + "C"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "C" + "B"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["B", "C"])  # slow
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["B", "C"])  # slow

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


def test_cap_extraction_common_emitter_reverse():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common emitter config PI network
    g3 = 1j * 2 * np.pi * freq * cbe
    g2 = 1j * 2 * np.pi * freq * cbc
    g1 = 0
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "B" + "B"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "C" + "C"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "B" + "C"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "C" + "B"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["C", "B"])  # slow
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["C", "B"])  # slow

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


def test_cap_extraction_common_collector():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common collector config PI network
    g1 = 1j * 2 * np.pi * freq * cbc
    g2 = 1j * 2 * np.pi * freq * cbe
    g3 = 1j * 2 * np.pi * freq * cce
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "B" + "B"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "E" + "E"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "B" + "E"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "E" + "B"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["B", "E"])  # slow
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["B", "E"])  # slow

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


def test_cap_extraction_common_collector_reversed():
    # create a test dataframe
    df = DataFrame()
    freq = np.logspace(3, 10, 101)
    cbe = 15e-15
    cbc = 10e-15
    cce = 0
    # common collector config PI network
    g3 = 1j * 2 * np.pi * freq * cbc
    g2 = 1j * 2 * np.pi * freq * cbe
    g1 = 1j * 2 * np.pi * freq * cce
    y_common_e = np.zeros((len(freq), 2, 2), dtype=np.complex128)
    y_common_e[:, 0, 0] = g1 + g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 0, 1] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 0] = -g2  # pylint: disable=unsupported-assignment-operation
    y_common_e[:, 1, 1] = g2 + g3  # pylint: disable=unsupported-assignment-operation
    df[specifiers.FREQUENCY] = freq
    df[specifiers.SS_PARA_Y + "B" + "B"] = y_common_e[:, 0, 0]
    df[specifiers.SS_PARA_Y + "E" + "E"] = y_common_e[:, 1, 1]
    df[specifiers.SS_PARA_Y + "B" + "E"] = y_common_e[:, 0, 1]
    df[specifiers.SS_PARA_Y + "E" + "B"] = y_common_e[:, 1, 0]
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["E", "B"])
    df.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["E", "B"])

    # extract cbe
    cbe_extract = df[specifiers.CAPACITANCE + "B" + "E"].to_numpy()
    cbe_extract = np.unique(cbe)
    assert np.isclose(cbe, cbe_extract)
    # extract cbc
    cbc_extract = df[specifiers.CAPACITANCE + "B" + "C"].to_numpy()
    cbc_extract = np.unique(cbc)
    assert np.isclose(cbc, cbc_extract)


if __name__ == "__main__":
    from DMT.core import Plot

    import cProfile, pstats, io

    profiler = cProfile.Profile()
    time_start = datetime.now()
    profiler.enable()
    test_cap_extraction_common_emitter()
    test_cap_extraction_common_emitter_reverse()
    test_cap_extraction_common_base()
    test_cap_extraction_common_base_reversed()
    test_cap_extraction_common_collector()
    test_cap_extraction_common_collector_reversed()
    df_deemb = test_data_processor_and_deem()
    profiler.disable()
    print("Runtime: " + str(datetime.now() - time_start))

    s = io.StringIO()
    ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")

    ps.print_stats(50)  # print the 50 methods with the longest cumulative time

    print(s.getvalue())
    # plot the result
    plt_cbe = Plot("C_BE(V_BE)", style="bw")
    for freq_a in [1e9, 5e9, 10e9]:
        df_tmp = df_deemb[df_deemb["FREQ"] == freq_a]
        plt_cbe.add_data_set(df_tmp["V_B"].to_numpy(), df_tmp["C_BE"].to_numpy(), label="meas")

    plt_cbe.plot_pyqtgraph(show=True)



"""Testing the VA-Files Tree
"""
import pytest
import base64, zlib
from pathlib import Path
from DMT.core import VAFile
from DMT.core.va_file import VACode

path_here = Path(__file__).resolve().parent
path_to_cmc_model = path_here / "test_va_code" / "diode_cmc_160823"


def test_files():
    with pytest.raises(IOError) as exec_info:
        root_error = VAFile("root.va", files={"leaf.va": VACode(code="test code")})

    va_file = VAFile(
        "diode_cmc.va",
        files={
            "diode_cmc.va": VACode(code="test code"),
            "DIODE_CMC_InitModel.include": VACode(code="test child_code 1"),
            "DIODE_CMC_macrodefs.include": VACode(code="test child_code 2"),
        },
    )

    assert set(va_file.files.keys()) == {
        "diode_cmc.va",
        "DIODE_CMC_InitModel.include",
        "DIODE_CMC_macrodefs.include",
    }
    assert va_file.files["DIODE_CMC_InitModel.include"] == "test child_code 1"


def test_structure_generation():
    va_file = VAFile("diode_cmc.va")
    va_file.read_structure(path_to_cmc_model)

    assert len(va_file) == 7

    assert va_file.files[va_file.root] == (path_to_cmc_model / "diode_cmc.va").read_text()
    assert (
        va_file.files["DIODE_CMC_parlist.include"]
        == (path_to_cmc_model / "DIODE_CMC_parlist.include").read_text()
    )


def test_code_compression():
    va_file = VAFile("diode_cmc.va")
    va_file.read_structure(path_to_cmc_model)

    code_compressed, crc = va_file.files[va_file.root].code_compressed

    assert len(code_compressed) < len(va_file.files[va_file.root].code)
    assert crc == 871444854

    # decompress and compare:
    code = base64.b85decode(code_compressed.encode("utf-8"))
    code = zlib.decompress(code).decode("utf-8")
    assert code == (path_to_cmc_model / "diode_cmc.va").read_text()


def test_tree_hash():
    va_file = VAFile("diode_cmc.va")
    va_file.read_structure(path_to_cmc_model)

    assert va_file.get_tree_hash() == "199d457ec02425336859dfbee42dd5fd"


def test_dict_export():
    va_file = VAFile("diode_cmc.va")
    va_file.read_structure(path_to_cmc_model)

    export_dict = va_file.export_dict()

    va_file_imported = VAFile.import_dict(export_dict)

    assert va_file_imported.get_tree_hash() == "199d457ec02425336859dfbee42dd5fd"
    assert (
        va_file_imported.files[va_file_imported.root]
        == (path_to_cmc_model / "diode_cmc.va").read_text()
    )
    assert (
        va_file_imported.files["DIODE_CMC_parlist.include"]
        == (path_to_cmc_model / "DIODE_CMC_parlist.include").read_text()
    )

    export_dict_compressed = va_file.export_dict(compressed_code=True)
    va_file_imported_compressed = VAFile.import_dict(export_dict_compressed)

    assert va_file_imported_compressed.get_tree_hash() == "199d457ec02425336859dfbee42dd5fd"
    assert (
        va_file_imported_compressed.files[va_file_imported_compressed.root]
        == (path_to_cmc_model / "diode_cmc.va").read_text()
    )
    assert (
        va_file_imported_compressed.files["DIODE_CMC_parlist.include"]
        == (path_to_cmc_model / "DIODE_CMC_parlist.include").read_text()
    )


def test_write_files():
    va_file = VAFile("diode_cmc.va")
    va_file.read_structure(path_to_cmc_model)

    target_path = path_here.parent / "tmp" / "test_write_vafile"
    va_file.write_files(target_path)

    assert (target_path / "DIODE_CMC_SIMKIT_macrodefs.include").is_file()

    va_file_read = VAFile("diode_cmc.va")
    va_file_read.read_structure(target_path)

    assert va_file == va_file_read  # compare


if __name__ == "__main__":
    test_files()
    test_structure_generation()
    test_code_compression()
    test_tree_hash()
    test_dict_export()
    test_write_files()



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
from DMT.core import __version__ as ver

# -- Project information -----------------------------------------------------

project = "DMT - Device Modeling Toolkit"
copyright = "2022, M. Mueller and M. Krattenmacher"
author = "M. Mueller and M. Krattenmacher"

# The short X.Y version
version = f"{ver.major}.{ver.minor}"
# The full version, including alpha/beta/rc tags
release = str(ver)


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.linkcode",
    "sphinx.ext.graphviz",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.pdfembed",
    "numpydoc",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = []

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

themedir = os.path.join(os.path.dirname(__file__), "_theme")
if not os.path.isdir(themedir):
    raise RuntimeError(
        "Get the scipy-sphinx-theme first, " "via git submodule init && git submodule update"
    )

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "edit_link": False,
}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [themedir]


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "../logo/DMT_Logo_wText.png"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%d.%m.%Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
html_search_language = "en"

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "DMTdoc"

numpydoc_show_class_members = False
numpydoc_class_members_toctree = False
# numpydoc_xref_param_typebool
#     Whether to create cross-references for the parameter types in the Parameters, Other Parameters, Returns and Yields sections of the docstring. False by default.
numpydoc_xref_param_type = True

numpydoc_xref_aliases = {
    "CircuitElement": "DMT.core.circuit.CircuitElement",
    "Circuit": "DMT.core.circuit.Circuit",
    "MCard": "DMT.core.mcard.MCard",
    "McParameterCollection": "DMT.core.mc_parameter.McParameterCollection",
    "McParameter": "DMT.core.mc_parameter.McParameter",
    "core": "DMT.core",
    "ngspice": "DMT.ngspice",
    "xyce": "DMT.xyce",
    "exceptions": "DMT.exceptions",
    "config": "DMT.config",
    "Hdev": "DMT.Hdev",
    "VerilogAE": "DMT.VerilogAE",
}

# intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "mne": ("https://mne.tools/dev", None),
    "numpy": ("https://www.numpy.org/devdocs", None),
    "scipy": ("https://scipy.github.io/devdocs", None),
    "matplotlib": ("https://matplotlib.org", None),
}
intersphinx_timeout = 10

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "DMT.tex", "DMT Documentation", "M. Mueller and M. Krattenmacher", "manual"),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "dmt", "DMT Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "DMT",
        "DMT Documentation",
        author,
        "DMT",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for linkcode extension ----------------------------------------------
def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"https://gitlab.com/dmt-development/dmt-core/-/tree/main/{filename:s}.py"


# -- Options for autodoc extension ----------------------------------------------

# -- autodoc_typehints_format
#     This value controls the format of typehints. The setting takes the following values:
#         'fully-qualified' – Show the module name and its name of typehints
#         'short' – Suppress the leading module names of the typehints (ex. io.StringIO -> StringIO) (default)
autodoc_typehints_format = "short"



import numpy as np
import pytest
from pathlib import Path
from DMT.core import specifiers, sub_specifiers, read_elpa, read_mdm, Sweep, get_sweepdef

folder_path = Path(__file__).resolve().parent


def test_errors():
    df_elpa = read_elpa(folder_path / "HBT_vbc.elpa")

    with pytest.raises(IOError):  # sweep is tried created from 2 forced currents
        _sweep = get_sweepdef(
            df_elpa,
            inner_sweep_voltage=specifiers.CURRENT + "B",
            outer_sweep_voltage=specifiers.CURRENT + "C",
        )

    df_elpa[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED] = df_elpa["VBE"]
    df_elpa[specifiers.VOLTAGE + "E" + sub_specifiers.FORCED] = df_elpa["VBE"]
    df_elpa[specifiers.VOLTAGE + "C" + sub_specifiers.FORCED] = df_elpa["VCE"]

    with pytest.raises(
        IOError
    ):  # sweeps need an entry point for the sweep tree -> potential with single value for all ops...
        _sweep = get_sweepdef(df_elpa)


def test_convert_elpa():
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_vce = specifiers.VOLTAGE + ["C", "E"]
    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe_forced = specifiers.VOLTAGE + ["B", "E"] + sub_specifiers.FORCED
    col_vce_forced = specifiers.VOLTAGE + ["C", "E"] + sub_specifiers.FORCED
    col_vbc_forced = specifiers.VOLTAGE + ["B", "C"] + sub_specifiers.FORCED

    df_elpa = read_elpa(folder_path / "HBT_vbc.elpa")

    df_elpa[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED] = df_elpa["VBE"]
    df_elpa[specifiers.VOLTAGE + "E" + sub_specifiers.FORCED] = df_elpa["VBE"] * 0
    df_elpa[specifiers.VOLTAGE + "C" + sub_specifiers.FORCED] = df_elpa["VCE"]

    sweep = Sweep.get_sweep_from_dataframe(df_elpa, temperature=300.0)

    df_sweep = sweep.create_df()

    df_elpa.ensure_specifier_column(col_vbe_forced)
    df_elpa.ensure_specifier_column(col_vbc_forced)
    df_elpa.ensure_specifier_column(col_vce_forced)
    df_sweep.ensure_specifier_column(col_vbe)
    df_sweep.ensure_specifier_column(col_vbc)
    df_sweep.ensure_specifier_column(col_vce)
    assert df_elpa.shape[0] == df_sweep.shape[0]
    assert all(
        [
            np.isclose(vbe_elpa, vbe_sweep)
            for (vbe_elpa, vbe_sweep) in zip(
                sorted(df_elpa[col_vbe_forced]), sorted(df_sweep[col_vbe])
            )
        ]
    )
    assert all(
        [
            np.isclose(vce_elpa, vce_sweep)
            for (vce_elpa, vce_sweep) in zip(
                sorted(df_elpa[col_vce_forced]), sorted(df_sweep[col_vce])
            )
        ]
    )
    assert all(
        [
            np.isclose(vbc_elpa, vbc_sweep)
            for (vbc_elpa, vbc_sweep) in zip(
                sorted(df_elpa[col_vbc_forced]), sorted(df_sweep[col_vbc])
            )
        ]
    )

    df_elpa = read_elpa(folder_path / "HBT_vce.elpa")

    df_elpa[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED] = df_elpa["VBE"]
    df_elpa[specifiers.VOLTAGE + "E" + sub_specifiers.FORCED] = df_elpa["VBE"] * 0
    df_elpa[specifiers.VOLTAGE + "C" + sub_specifiers.FORCED] = df_elpa["VCE"]

    sweep = Sweep.get_sweep_from_dataframe(df_elpa, temperature=300.0)

    df_sweep = sweep.create_df()

    df_elpa.ensure_specifier_column(col_vbe_forced)
    df_elpa.ensure_specifier_column(col_vbc_forced)
    df_elpa.ensure_specifier_column(col_vce_forced)
    df_sweep.ensure_specifier_column(col_vbe)
    df_sweep.ensure_specifier_column(col_vbc)
    df_sweep.ensure_specifier_column(col_vce)
    assert df_elpa.shape[0] == df_sweep.shape[0]
    assert all(
        [
            np.isclose(vbe_elpa, vbe_sweep)
            for (vbe_elpa, vbe_sweep) in zip(
                sorted(df_elpa[col_vbe_forced]), sorted(df_sweep[col_vbe])
            )
        ]
    )
    assert all(
        [
            np.isclose(vbc_elpa, vbc_sweep)
            for (vbc_elpa, vbc_sweep) in zip(
                sorted(df_elpa[col_vbc_forced]), sorted(df_sweep[col_vbc])
            )
        ]
    )
    assert all(
        [
            np.isclose(vce_elpa, vce_sweep)
            for (vce_elpa, vce_sweep) in zip(
                sorted(df_elpa[col_vce_forced]), sorted(df_sweep[col_vce])
            )
        ]
    )


def test_convert_mdm():
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_vce = specifiers.VOLTAGE + ["C", "E"]
    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe_forced = specifiers.VOLTAGE + ["B", "E"] + sub_specifiers.FORCED
    col_vce_forced = specifiers.VOLTAGE + ["C", "E"] + sub_specifiers.FORCED
    col_vbc_forced = specifiers.VOLTAGE + ["B", "C"] + sub_specifiers.FORCED
    # load test measurements
    df_mdm = read_mdm(folder_path / "test_data" / "Spar_vb.mdm")

    # correct data format of measurements
    nodes = ["B", "C", "E", "S"]
    df_mdm = df_mdm.clean_data(
        nodes,
        "E",
        fallback={
            "S_DEEMB(1,1)": None,
            "S_DEEMB(2,1)": None,
            "S_DEEMB(1,2)": None,
            "S_DEEMB(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_mdm[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED] = df_mdm[specifiers.VOLTAGE + "B"]
    df_mdm[specifiers.VOLTAGE + "E" + sub_specifiers.FORCED] = df_mdm[specifiers.VOLTAGE + "E"]
    df_mdm[specifiers.VOLTAGE + "C" + sub_specifiers.FORCED] = df_mdm[specifiers.VOLTAGE + "C"]

    sweep = Sweep.get_sweep_from_dataframe(df_mdm, temperature=300.0)

    df_sweep = sweep.create_df()

    df_mdm.ensure_specifier_column(col_vbe_forced)
    df_mdm.ensure_specifier_column(col_vbc_forced)
    df_mdm.ensure_specifier_column(col_vce_forced)
    df_sweep.ensure_specifier_column(col_vbe)
    df_sweep.ensure_specifier_column(col_vbc)
    df_sweep.ensure_specifier_column(col_vce)

    assert df_mdm.shape[0] == df_sweep.shape[0]
    assert all(
        [
            np.isclose(vbe_mdm, vbe_sweep)
            for (vbe_mdm, vbe_sweep) in zip(
                sorted(df_mdm[col_vbe_forced]), sorted(df_sweep[col_vbe])
            )
        ]
    )
    assert all(
        [
            np.isclose(vbc_mdm, vbc_sweep)
            for (vbc_mdm, vbc_sweep) in zip(
                sorted(df_mdm[col_vbc_forced]), sorted(df_sweep[col_vbc])
            )
        ]
    )
    assert all(
        [
            np.isclose(vce_mdm, vce_sweep)
            for (vce_mdm, vce_sweep) in zip(
                sorted(df_mdm[col_vce_forced]), sorted(df_sweep[col_vce])
            )
        ]
    )
    assert all(
        [
            np.isclose(freq_mdm, freq_sweep)
            for (freq_mdm, freq_sweep) in zip(
                sorted(df_mdm[specifiers.FREQUENCY]), sorted(df_sweep[specifiers.FREQUENCY])
            )
        ]
    )


if __name__ == "__main__":
    test_errors()
    test_convert_elpa()
    test_convert_mdm()



""" test ngspice input file generation.
"""

import numpy as np
import logging
from pathlib import Path
from DMT.core import SimCon, Plot, DutType, Sweep, specifiers
from DMT.ngspice import DutNgspice
from DMT.ADS import DutAds
from DMT.hl2 import McHicum, Hl2Model, VA_FILES

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ADS.log",
    filemode="w",
)


def compare_ngspice_ads():

    mc = McHicum(load_model_from_path=folder_path / "npn_full_sh.lib")

    ###### Disable branches
    ## General settings
    mc.set_values(
        {
            "flcomp": 2.4,
            "flnqs": 0,
            "flsh": 0,
            # 'ibeis' : 1e-10,
            "c10": 1e-30,
            "re": 3,
            "rcx": 3,
            "rbx": 1,
            "rbi0": 3,
            "t0": 1e-13,
            "ireps": 1e-13,
            "ireis": 1e-13,
            "cjci0": 1e-15,
            "hjei": 2,
            "ahjei": 1,
            "tbhrec": 1e-12,  # has not much influence?
        }
    )

    # flags to turn off elements
    re = True
    rcx = True
    rbx = True
    rbi0 = True
    ibcxs = True
    ibcis = True
    ibeps = True
    ireps = True
    ibeis = True
    ireis = True
    ibets = True
    tbhrec = True
    cbcpar = True
    cjcx0 = True
    cjci0 = True
    cbepar = True
    cjep0 = True
    cjei0 = True
    cjs0 = False  # not implemented
    t0 = True
    ick = True
    it = True
    iavl = False

    if not it:
        mc.set_values({"c10": 0})  # transfer current
    if not iavl:
        mc.set_values({"favl": 0.0, "qavl": 0.0})  # avalanche current

    ## Resistances
    if not re:
        mc.set_values({"re": 0})  # emitter resistance
    if not rcx:
        mc.set_values({"rcx": 0})  # collector resistance
    if not rbx:
        mc.set_values({"rbx": 0})  # external base resistance
    if not rbi0:
        mc.set_values({"rbi0": 0})  # internal base resistance

    ## Diodes
    if not ibcxs:
        mc.set_values({"ibcxs": 0})  # external collector diode
    if not ibcis:
        mc.set_values({"ibcis": 0})  # interal collector diode
    if not ibeps:
        mc.set_values({"ibeps": 0})  # peripheral emitter injection diode
    if not ireps:
        mc.set_values({"ireps": 0})  # peripheral emitter recombinatin diode
    if not ibeis:
        mc.set_values({"ibeis": 0})  # internal emitter injection diode
    if not ireis:
        mc.set_values({"ireis": 0})  # internal emitter recombinatin diode
    if not ibets:
        mc.set_values({"ibets": 0})  # internal emitter tunneling diode
    if not tbhrec:
        mc.set_values({"tbhrec": 0})  # increased recombination

    ## Capacitances
    if not cbcpar:
        mc.set_values({"cbcpar": 0})  # parasitic collector capacitance
    if not cbepar:
        mc.set_values({"cbepar": 0})  # parasitic collector capacitance
    if not cjcx0:
        mc.set_values({"cjcx0": 0})  # external collector junction capacitance
    if not cjci0:
        mc.set_values({"cjci0": 0})  # parasitic emitter capacitance
    if not cjep0:
        mc.set_values({"cjep0": 0})  # peripheral emitter capacitance
    if not cjei0:
        mc.set_values({"cjei0": 0})  # internal emitter capacitance
    if not cjs0:
        mc.set_values({"cjs0": 0})  # substrate junction capacitance

    ## diffusion charge effects
    if not ick:
        mc.set_values(
            {"rci0": 1e-9}
        )  # collector resistance -> increases ICK to inf -> no high current charge
    if not t0:
        mc.set_values(
            {"t0": 1e-15, "dt0h": 0.0, "tbvl": 0.0}
        )  # zero bias transit time -> no charge storage at all

    ######

    mc.va_file = VA_FILES["L2V2.4.0_internal"]
    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file

    # create a sweep with ONE operating point
    sweepdef = [
        # {'var_name':specifiers.FREQUENCY    ,  'sweep_order':4, 'sweep_type':'LOG' , 'value_def':[1,2,11]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.85]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[-2.0]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0.0, 1.0, 10],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [2.0],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    sim_con = SimCon()

    # first simulate ADS
    sim_con.append_simulation(dut=dut_ads, sweep=sweep)
    sim_con.run_and_read(force=False)

    # read initial conditions
    data_ads = dut_ads.get_data(sweep=sweep)
    ve = 0
    vb = data_ads["VBE"].to_numpy()[0]
    vc = data_ads["VCE"].to_numpy()[0]
    vbi = data_ads["VBiE"].to_numpy()[0]
    vbp = data_ads["VBpE"].to_numpy()[0]
    vci = data_ads["VCiE"].to_numpy()[0]
    vei = data_ads["VEiE"].to_numpy()[0]
    initial_conditions = {
        "icVE": ve * 1.1,
        "icVEi": vei * 1.1,
        "icVC": vc * 1.1,
        "icVCi": vci * 1.1,
        "icVB": vb * 1.1,
        "icVBi": vbi * 1.1,
        "icVBp": vbp * 1.1,
    }

    # simulate ngspice using initial condition from ADS
    dut_ngspice = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        initial_conditions=initial_conditions,
        simulator_options={"GMIN": 1e-18},
    )
    sim_con.append_simulation(dut=dut_ngspice, sweep=sweep)
    sim_con.run_and_read(force=True)

    data_ngspice = dut_ngspice.get_data(sweep=sweep)
    model = Hl2Model(model_name="hl2", version="2.4.3")
    kwargs = mc.to_kwargs()

    # ngspice operating point
    vbiei = data_ngspice["VBIEI"].to_numpy()
    vbici = data_ngspice["VBICI"].to_numpy()
    vciei = data_ngspice["VCIEI"].to_numpy()
    temp = data_ngspice["TK"].to_numpy()
    # ngspice currents
    ibei_ngspice = data_ngspice["IBEI"].to_numpy()
    ibei_ads = data_ads["IBIEI"].to_numpy()
    it_ngspice = data_ngspice["IT"].to_numpy()
    it_ads = data_ads["IT_"].to_numpy()
    # some assertions since I am at this point clueeless ... ##########################################
    # assert np.allclose(
    #     model.model_ibei_temp(vbe=vbiei, t_dev=temp,**kwargs),
    #     data_ngspice['IBEI'].to_numpy(),
    #     rtol=1e-2
    # )
    try:
        assert np.allclose(
            model.model_ibci_temp(vbc=vbici, t_dev=temp, **kwargs),
            data_ngspice["IBCI"].to_numpy(),
            rtol=1e-3,
        )  # not exactly zero due to gmin
    except AssertionError:
        pass  # gmin makes this assertion fail for small current
    assert np.allclose(
        model.model_ick_temp(vce=vciei, t_dev=temp, **kwargs),
        data_ngspice["ICK"].to_numpy(),
        rtol=1e-3,
    )  # not exactly zero due to gmin
    assert np.allclose(
        model.model_cjci_temp(vbc=vbici, t_dev=temp, **kwargs),
        data_ngspice["CMUI"].to_numpy(),
        rtol=1e-3,
    )
    # assert np.allclose(
    #     model.model_cjei_temp(vbe=vbiei, t_dev=temp, **kwargs),
    #     data_ngspice['CPII'].to_numpy(),
    #     rtol=1e-3
    # ) # does not work because of CdE

    # using currents

    # assert np.allclose(
    #     ibei_ngspice,
    #     ibei_ads,
    #     rtol=1e-3
    # )
    # assert np.allclose(
    #     it_ngspice,
    #     it_ads,
    #     rtol=1e-3
    # )

    return data_ngspice, data_ads


if __name__ == "__main__":
    data_ngspice, data_ads = compare_ngspice_ads()
    # define plots
    plt_gummel = Plot(
        r"$I(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{C}},I_{\mathrm{B}}$",
        y_log=True,
        style="markers_lines",
    )
    plt_ibei = Plot(
        r"$IBEI(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"IBEI",
        y_log=True,
        style="markers_lines",
    )
    plt_ibci = Plot(
        r"$IBCI(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"IBCI",
        y_log=True,
        style="markers_lines",
    )
    plt_it = Plot(
        r"$IT(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"IT",
        y_log=True,
        style="markers_lines",
    )
    # end define plots

    # fill plots
    # gummel
    plt_gummel.add_data_set(data_ngspice["V_B"], data_ngspice["I_C"], label="ngspice $I_C$")
    plt_gummel.add_data_set(data_ads["V_B"], data_ads["I_C"], label="ADS $I_C$")
    plt_gummel.add_data_set(data_ngspice["V_B"], data_ngspice["I_B"], label="ngspice $I_B$")
    plt_gummel.add_data_set(data_ads["V_B"], data_ads["I_B"], label="ADS $I_B$")
    # ibei
    plt_ibei.add_data_set(data_ngspice["V_B"], data_ngspice["IBEI"], label="ngspice ibei")
    plt_ibei.add_data_set(data_ads["V_B"], data_ads["IBIEI"], label="ADS ibei")
    # ibci
    plt_ibci.add_data_set(data_ngspice["V_B"], data_ngspice["IBCI"], label="ngspice ibci")
    plt_ibci.add_data_set(data_ads["V_B"], data_ads["IBICI"], label="ADS ibci")
    # it
    plt_it.add_data_set(data_ngspice["V_B"], data_ngspice["IT"], label="ngspice it")
    plt_it.add_data_set(data_ads["V_B"], data_ads["IT_"], label="ADS it")

    plt_it.plot_pyqtgraph(show=False)
    plt_ibci.plot_pyqtgraph(show=False)
    plt_ibei.plot_pyqtgraph(show=False)
    plt_gummel.plot_pyqtgraph(show=True)



"""
This example shows how to read in measurement data in batch mode.
"""
from pathlib import Path
from DMT.core import DutLib, DutMeas, DutType

# path to DMT test cases
path_test = Path(__file__).resolve().parent.parent.parent.parent / "test"


def dut_filter(dut_name):
    """This function is given to the "import_directory" function as an argument. It will be called for all valid folders. If
    it returns a DutMeas object, all data will be read-in and assigned to the DutMeas object..
    """
    if "0p25x10" in dut_name:
        return DutMeas(
            path_test / "tmp",  # here the device is stored when its save() routine is called
            "CBEBC_0p25x10",
            DutType.npn,
            wafer="xy",
            die="z",
            reference_node="E",
        )
    else:
        return None


# the device and its measurement data are in the folder:
#'test/test_core_no_interfaces/test_data/0p25x10/'
#'test/test_core_no_interfaces/test_data/0p25x10/meas_1'
#'test/test_core_no_interfaces/test_data/0p25x10/meas_2'
#'test/test_core_no_interfaces/test_data/0p25x10/...'

# create a library
lib = DutLib()

# read in all data for all devices
# the folders that contain measurement data for one device structure lie one level below the given first argument.
lib.import_directory(path_test / "test_core_no_interfaces" / "test_data", dut_filter, dut_level=1)

# now the data has been read-in, but the data columns are not yet ensured to
# follow the DMT specifiers grammar for electrical variables => important to have consistent data within DMT

# data that should be deleted is given as a dictionary:
# (every measurement folk has his own naming conventions -> standardize them here)
fallback = {
    "S_DEEMB(1,1)": None,
    "S_DEEMB(1,2)": None,
    "S_DEEMB(2,1)": None,
    "S_DEEMB(2,2)": None,
}
for dut in lib:
    # now ensure the data is in DMT format!
    dut.clean_data(fallback=fallback)


# lib.save() can be now be called to store all the data in HDF5 format => memory efficient and fast.
